import logging
import os
import time
import traceback
from datetime import datetime
from logging import Logger
from pathlib import Path
from typing import Optional

from bauh.api.paths import CACHE_DIR
from bauh.commons.system import ProcessHandler

SYNC_FILE = f'{CACHE_DIR}/arch/db_sync'


def should_sync(arch_config: dict, aur_supported: bool, handler: Optional[ProcessHandler], logger: logging.Logger):
    if aur_supported or arch_config['repositories']:
        if os.path.exists(SYNC_FILE):
            with open(SYNC_FILE) as f:
                sync_file = f.read()

            try:
                sync_time = datetime.fromtimestamp(int(sync_file))
                now = datetime.now()

                if now > sync_time and now.day != sync_time.day:
                    logger.info("Package databases synchronization out of date")
                else:
                    msg = "Package databases already synchronized"
                    logger.info(msg)
                    if handler:
                        handler.watcher.print(msg)
                    return False
            except Exception:
                logger.warning("Could not convert the database synchronization time from '{}".format(SYNC_FILE))
                traceback.print_exc()
        return True
    else:
        msg = "Package databases synchronization disabled"
        if handler:
            handler.watcher.print(msg)
        logger.info(msg)
        return False


def register_sync(logger: Logger):
    try:
        Path('/'.join(SYNC_FILE.split('/')[0:-1])).mkdir(parents=True, exist_ok=True)
        with open(SYNC_FILE, 'w+') as f:
            f.write(str(int(time.time())))
    except Exception:
        logger.error("Could not write to database sync file '{}'".format(SYNC_FILE))
        traceback.print_exc()



from logging import Logger
from typing import Tuple

from PyQt5.QtCore import QObject
from PyQt5.QtWidgets import QApplication

from bauh.context import new_qt_application
from bauh.view.qt.systray import TrayIcon


def new_tray_icon(app_config: dict, logger: Logger) -> Tuple[QApplication, QObject]:
    app = new_qt_application(app_config=app_config, logger=logger, quit_on_last_closed=True)
    tray_icon = TrayIcon(screen_size=app.primaryScreen().size(), config=app_config, logger=logger)
    tray_icon.show()

    return app, tray_icon



from abc import ABC, abstractmethod
from typing import Set, Optional


class MemoryCache(ABC):
    """
    Represents a memory cache.
    """

    @abstractmethod
    def is_enabled(self):
        pass

    @abstractmethod
    def add(self, key: str, val: object):
        pass

    @abstractmethod
    def add_non_existing(self, key: str, val: object):
        pass

    @abstractmethod
    def get(self, key: str):
        pass

    @abstractmethod
    def delete(self, key):
        pass

    @abstractmethod
    def keys(self) -> Set[str]:
        pass

    @abstractmethod
    def clean_expired(self):
        pass


class MemoryCacheFactory(ABC):
    """
    Instantiate new memory cache instances.
    """

    @abstractmethod
    def new(self, expiration: Optional[int]) -> MemoryCache:
        """
        :param expiration: expiration time for the cache keys in seconds. Use -1 to disable this feature.
        :return:
        """
        pass



APP_ATTRS = ('name', 'description', 'github', 'source', 'version', 'url_download', 'url_icon', 'url_screenshot', 'license', 'author', 'categories')
RELEASE_ATTRS = ('version', 'url_download', 'published_at')


SEARCH_APPS_BY_NAME_OR_DESCRIPTION = "SELECT {} FROM apps".format(','.join(APP_ATTRS)) + " WHERE lower(name) LIKE '%{}%' or lower(description) LIKE '%{}%'"
FIND_APP_ID_BY_NAME_AND_GITHUB = "SELECT id FROM apps WHERE lower(name) = '{}' and lower(github) = '{}'"
FIND_APPS_BY_NAME = "SELECT name, github, version, url_download FROM apps WHERE lower(name) IN ({})"
FIND_APPS_BY_NAME_ONLY_NAME = "SELECT name FROM apps WHERE lower(name) IN ({})"
FIND_APPS_BY_NAME_FULL = "SELECT {} FROM apps".format(','.join(APP_ATTRS)) + " WHERE lower(name) IN ({})"
FIND_RELEASES_BY_APP_ID = "SELECT {} FROM releases".format(','.join(RELEASE_ATTRS)) + " WHERE app_id = {}"



import os
from typing import Optional

from bauh import __app_name__
from bauh.api.paths import CONFIG_DIR, TEMP_DIR, CACHE_DIR, get_temp_dir
from bauh.commons import resource

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
ARCH_CACHE_DIR = f'{CACHE_DIR}/arch'
CATEGORIES_FILE_PATH = f'{ARCH_CACHE_DIR}/categories.txt'
URL_CATEGORIES_FILE = f'https://raw.githubusercontent.com/vinifmor/{__app_name__}-files/master/arch/categories.txt'
URL_GPG_SERVERS = f'https://raw.githubusercontent.com/vinifmor/{__app_name__}-files/master/arch/gpgservers.txt'
ARCH_CONFIG_DIR = f'{CONFIG_DIR}/arch'
CUSTOM_MAKEPKG_FILE = f'{ARCH_CONFIG_DIR}/makepkg.conf'
AUR_INDEX_FILE = f'{ARCH_CACHE_DIR}/aur/index.txt'
AUR_INDEX_TS_FILE = f'{ARCH_CACHE_DIR}/aur/index.ts'
CONFIG_FILE = f'{CONFIG_DIR}/arch.yml'
UPDATES_IGNORED_FILE = f'{ARCH_CONFIG_DIR}/updates_ignored.txt'
EDITABLE_PKGBUILDS_FILE = f'{ARCH_CONFIG_DIR}/aur/editable_pkgbuilds.txt'
IGNORED_REBUILD_CHECK_FILE = f'{ARCH_CONFIG_DIR}/aur/ignored_rebuild_check.txt'


def get_pkgbuild_dir(user: Optional[str] = None) -> str:
    return f'{get_temp_dir(user) if user else TEMP_DIR}/arch'


def get_icon_path() -> str:
    return resource.get_path('img/arch.svg', ROOT_DIR)


def get_repo_icon_path() -> str:
    return resource.get_path('img/repo.svg', ROOT_DIR)



from typing import Optional, Any


class Value:

    def __init__(self, value: Optional[Any] = None):
        self.value = value

    def __repr__(self):
        return str(self.value)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        if isinstance(other, Value):
            return self.value == other.value

    def __hash__(self):
        return hash(self.value)



import os
import traceback
from datetime import datetime, timedelta
from logging import Logger
from pathlib import Path
from threading import Thread
from typing import Optional, Dict

from bauh.api.abstract.handler import TaskManager
from bauh.api.abstract.model import SuggestionPriority
from bauh.api.http import HttpClient
from bauh.commons.suggestions import parse
from bauh.gems.debian import DEBIAN_ICON_PATH, DEBIAN_CACHE_DIR
from bauh.view.util.translation import I18n


class DebianSuggestionsDownloader(Thread):

    _file_suggestions: Optional[str] = None
    _file_suggestions_ts: Optional[str] = None
    _url_suggestions: Optional[str] = None

    @classmethod
    def file_suggestions(cls) -> str:
        if cls._file_suggestions is None:
            cls._file_suggestions = f'{DEBIAN_CACHE_DIR}/suggestions.txt'

        return cls._file_suggestions

    @classmethod
    def file_suggestions_timestamp(cls) -> str:
        if cls._file_suggestions_ts is None:
            cls._file_suggestions_ts = f'{cls.file_suggestions()}.ts'

        return cls._file_suggestions_ts

    def __init__(self, logger: Logger, http_client: HttpClient, i18n: I18n, file_url: Optional[str]):
        super(DebianSuggestionsDownloader, self).__init__()
        self._log = logger
        self.i18n = i18n
        self.http_client = http_client
        self._taskman: Optional[TaskManager] = None

        if file_url:
            self._file_url = file_url
        else:
            self._file_url = 'https://raw.githubusercontent.com/vinifmor/bauh-files/master/debian/suggestions_v1.txt'

        self.task_id = 'debian.suggs'

    def register_task(self, taskman: Optional[TaskManager]):
        self._taskman = taskman
        if taskman:
            self._taskman.register_task(id_=self.task_id, label=self.i18n['task.download_suggestions'],
                                        icon_path=DEBIAN_ICON_PATH)

    def is_local_suggestions_file(self) -> bool:
        return self._file_url and self._file_url.startswith('/')

    @property
    def taskman(self) -> TaskManager:
        if self._taskman is None:
            self._taskman = TaskManager()

        return self._taskman

    def should_download(self, debian_config: dict, only_positive_exp: bool = False) -> bool:
        if not self._file_url:
            self._log.error("No Debian suggestions file URL defined")
            return False

        if self.is_local_suggestions_file():
            return False

        try:
            exp_hours = int(debian_config['suggestions.exp'])
        except ValueError:
            self._log.error(f"The Debian configuration property 'suggestions.expiration' has a non int value set: "
                            f"{debian_config['suggestions']['expiration']}")
            return not only_positive_exp

        if exp_hours <= 0:
            self._log.info("Suggestions cache is disabled")
            return not only_positive_exp

        if not os.path.exists(self.file_suggestions()):
            self._log.info(f"'{self.file_suggestions()}' not found. It must be downloaded")
            return True

        if not os.path.exists(self.file_suggestions()):
            self._log.info(f"'{self.file_suggestions()}' not found. The suggestions file must be downloaded.")
            return True

        with open(self.file_suggestions_timestamp()) as f:
            timestamp_str = f.read()

        try:
            suggestions_timestamp = datetime.fromtimestamp(float(timestamp_str))
        except Exception:
            self._log.error(f'Could not parse the Debian cached suggestions timestamp: {timestamp_str}')
            traceback.print_exc()
            return True

        update = suggestions_timestamp + timedelta(hours=exp_hours) <= datetime.utcnow()
        return update

    def _save(self, text: str, timestamp: float):
        if not self._file_url:
            self._log.error("No Debian suggestions file URL defined")
            return

        if self.is_local_suggestions_file():
            return False

        self._log.info(f"Caching suggestions to '{self.file_suggestions()}'")

        cache_dir = os.path.dirname(self.file_suggestions())

        try:
            Path(cache_dir).mkdir(parents=True, exist_ok=True)
            cache_dir_ok = True
        except OSError:
            self._log.error(f"Could not create cache directory '{cache_dir}'")
            traceback.print_exc()
            cache_dir_ok = False

        if cache_dir_ok:
            try:
                with open(self.file_suggestions(), 'w+') as f:
                    f.write(text)
            except Exception:
                self._log.error(f"An exception happened while writing the file '{self.file_suggestions()}'")
                traceback.print_exc()

            try:
                with open(self.file_suggestions_timestamp(), 'w+') as f:
                    f.write(str(timestamp))
            except Exception:
                self._log.error(f"An exception happened while writing the file '{self.file_suggestions_timestamp()}'")
                traceback.print_exc()

    def read_cached(self) -> Optional[Dict[str, SuggestionPriority]]:
        if not self._file_url:
            self._log.error("No Debian suggestions file URL defined")
            return

        if self.is_local_suggestions_file():
            file_path, log_ref = self._file_url, 'local'
        else:
            file_path, log_ref = self.file_suggestions(), 'cached'

        self._log.info(f"Reading {log_ref} suggestions file {file_path}")

        try:
            with open(file_path) as f:
                sugs_str = f.read()
        except FileNotFoundError:
            self._log.warning(f"The {log_ref} suggestions file does not exist ({file_path})")
            return
        except OSError:
            self._log.warning(f"Could not read from the {log_ref} suggestions file ({file_path})")
            traceback.print_exc()
            return

        if not sugs_str:
            self._log.warning(f"The {log_ref} suggestions file '{file_path}' is empty")
            return

        return parse(sugs_str, self._log, 'Debian')

    def download(self) -> Optional[Dict[str, SuggestionPriority]]:
        if not self._file_url:
            self._log.error("No Debian suggestions file URL defined")
            return

        self.taskman.update_progress(self.task_id, progress=1, substatus=None)

        self._log.info(f"Downloading Debian suggestions from {self._file_url}")
        res = self.http_client.get(self._file_url)

        suggestions = None
        if res.status_code == 200 and res.text:
            self.taskman.update_progress(self.task_id, progress=50, substatus=None)
            suggestions = parse(res.text, self._log, 'Debian')

            if suggestions:
                self._save(text=res.text, timestamp=datetime.utcnow().timestamp())
            else:
                self._log.warning("No Debian suggestions to cache")
        else:
            self._log.warning(f"Could not retrieve Debian suggestions. "
                              f"Response (status={res.status_code}, text={res.text})")

        self.taskman.update_progress(self.task_id, progress=100, substatus=None)
        self.taskman.finish_task(self.task_id)
        return suggestions

    def read(self, debian_config: dict) -> Optional[Dict[str, int]]:
        if not self._file_url:
            self._log.error("No Debian suggestions file URL defined")
            return

        if self.is_local_suggestions_file() or not self.should_download(debian_config=debian_config):
            return self.read_cached()

        return self.download()

    def run(self):
        self.download()



import os
import shutil
import subprocess
import sys
import traceback
from typing import List, Tuple

from PyQt5.QtCore import QCoreApplication
from PyQt5.QtGui import QIcon
from colorama import Fore

from bauh import __app_name__
from bauh.api.abstract.controller import SoftwareManager
from bauh.api.paths import CONFIG_DIR, CACHE_DIR, TEMP_DIR
from bauh.commons.system import run_cmd
from bauh.view.util import resource


def notify_user(msg: str, icon_path: str = None):
    icon_id = icon_path

    if not icon_id:
        icon_id = get_default_icon()[0]

    os.system("notify-send -a {} {} '{}'".format(__app_name__, "-i {}".format(icon_id) if icon_id else '', msg))


def get_default_icon(system: bool = True) -> Tuple[str, QIcon]:
    if system:
        system_icon = QIcon.fromTheme(__app_name__)
        if not system_icon.isNull():
            return system_icon.name(), system_icon

    path = resource.get_path('img/logo.svg')
    return path, QIcon(path)


def restart_app():
    appimage_path = os.getenv('APPIMAGE')

    restart_cmd = [appimage_path] if appimage_path else [sys.executable, *sys.argv]

    subprocess.Popen(restart_cmd)
    QCoreApplication.exit()


def get_distro():
    if os.path.exists('/etc/arch-release'):
        return 'arch'

    if os.path.exists('/proc/version'):
        if 'ubuntu' in run_cmd('cat /proc/version').lower():
            return 'ubuntu'

    return 'unknown'


def clean_app_files(managers: List[SoftwareManager], logs: bool = True):

    if logs:
        print('[bauh] Cleaning configuration and cache files')

    for path in (CACHE_DIR, CONFIG_DIR, TEMP_DIR):
        if logs:
            print('[bauh] Deleting directory {}'.format(path))

        if os.path.exists(path):
            try:
                shutil.rmtree(path)
                if logs:
                    print('{}[bauh] Directory {} deleted{}'.format(Fore.YELLOW, path, Fore.RESET))
            except Exception:
                if logs:
                    print('{}[bauh] An exception has happened when deleting {}{}'.format(Fore.RED, path, Fore.RESET))
                    traceback.print_exc()

    if managers:
        for m in managers:
            m.clear_data()

    if logs:
        print('[bauh] Cleaning finished')



import os
import traceback
from datetime import datetime, timedelta
from logging import Logger
from pathlib import Path
from typing import Optional

import requests
import yaml

from bauh.api.http import HttpClient
from bauh.commons.util import map_timestamp_file
from bauh.gems.web import WEB_CACHE_DIR
from bauh.view.util.translation import I18n


class SuggestionsManager:

    def __init__(self, http_client: HttpClient, logger: Logger, i18n: I18n, file_url: Optional[str]):
        self.http_client = http_client
        self.logger = logger
        self.i18n = i18n
        if file_url:
            self._file_url = file_url
        else:
            self._file_url = "https://raw.githubusercontent.com/vinifmor/bauh-files/master/web/env/v2/suggestions.yml"
        self._cached_file_path = f'{WEB_CACHE_DIR}/suggestions.yml'
        self._cached_file_ts_path = map_timestamp_file(self._cached_file_path)

    @property
    def file_url(self) -> Optional[str]:
        return self._file_url

    def is_custom_local_file_mapped(self) -> bool:
        return self._file_url and self._file_url.startswith('/')

    def should_download(self, web_config: dict) -> bool:
        if not self._file_url:
            return False

        if self.is_custom_local_file_mapped():
            return False

        exp = web_config['suggestions']['cache_exp']
        if web_config['suggestions']['cache_exp'] is None:
            self.logger.info("No cache expiration defined for suggestions")
            return True

        try:
            exp = int(exp)
        except ValueError:
            self.logger.error(f"Error while parsing the 'suggestions.cache_exp' ({exp}) settings property")
            return True

        if exp <= 0:
            self.logger.info(f"No cache expiration defined for suggestions ({exp})")
            return True

        if not os.path.exists(self._cached_file_path):
            self.logger.info(f"No suggestions cached file found '{self._cached_file_path}'")
            return True

        if not os.path.exists(self._cached_file_ts_path):
            self.logger.info(f"No suggestions cache file timestamp found '{self._cached_file_ts_path}'")
            return True

        with open(self._cached_file_ts_path) as f:
            timestamp_str = f.read()

        try:
            sugs_timestamp = datetime.fromtimestamp(float(timestamp_str))
        except Exception:
            self.logger.error(f"Could not parse the cached suggestions file timestamp: {timestamp_str}")
            return True

        expired = sugs_timestamp + timedelta(days=exp) <= datetime.utcnow()

        if expired:
            self.logger.info("Cached suggestions file has expired.")
            return True
        else:
            self.logger.info("Cached suggestions file is up to date")
            return False

    def get_cached_file_path(self) -> str:
        return self._file_url if self.is_custom_local_file_mapped() else self._cached_file_path

    def read_cached(self, check_file: bool = True) -> dict:
        if self.is_custom_local_file_mapped():
            file_path, log_ref = self._file_url, 'local'
        else:
            file_path, log_ref = self._cached_file_path, 'cached'

        if check_file and not os.path.exists(file_path):
            self.logger.warning(f"{log_ref.capitalize()} suggestions file does not exist ({file_path})")
            return {}

        self.logger.info(f"Reading {log_ref} suggestions file '{file_path}'")
        with open(file_path) as f:
            sugs_str = f.read()

        if not sugs_str:
            self.logger.warning(f"{log_ref.capitalize()} suggestions file '{file_path}' is empty")
            return {}

        try:
            return yaml.safe_load(sugs_str)
        except Exception:
            self.logger.error("An unexpected exception happened")
            traceback.print_exc()
            return {}

    def download(self) -> dict:
        self.logger.info(f"Reading suggestions from {self._file_url}")
        try:
            suggestions = self.http_client.get_yaml(self._file_url, session=False)

            if suggestions:
                self.logger.info(f"{len(suggestions)} suggestions successfully read")
            else:
                self.logger.warning(f"Could not read suggestions from {self._file_url}")

        except (requests.exceptions.ConnectionError, requests.exceptions.ConnectTimeout):
            self.logger.warning("Internet seems to be off: it was not possible to retrieve the suggestions")
            suggestions = {}

        self.logger.info("Finished")
        return suggestions

    def save_to_disk(self, suggestions: dict, timestamp: float):
        if not suggestions:
            return

        if self.is_custom_local_file_mapped():
            return

        self.logger.info(f'Caching {len(suggestions)} suggestions to the disk')
        suggestions_file_dir = os.path.dirname(self._cached_file_path)

        try:
            Path(suggestions_file_dir).mkdir(parents=True, exist_ok=True)
        except OSError:
            self.logger.error(f"Could not generate the directory {suggestions_file_dir}")
            traceback.print_exc()
            return

        try:
            with open(self._cached_file_path, 'w+') as f:
                f.write(yaml.safe_dump(suggestions))
        except Exception:
            self.logger.error(f"Could write to {self._cached_file_path}")
            traceback.print_exc()
            return

        self.logger.info(f"{len(suggestions)} suggestions successfully cached to file '{self._cached_file_path}'")

        try:
            with open(self._cached_file_ts_path, 'w+') as f:
                f.write(str(timestamp))
        except Exception:
            self.logger.error(f"Could not write to {self._cached_file_ts_path}")
            traceback.print_exc()
            return

        self.logger.info(f"Suggestions cached file timestamp ({timestamp}) "
                         f"successfully saved at '{self._cached_file_ts_path}'")



import glob
import logging
import os
import time
import traceback
from threading import Lock, Thread
from typing import List, Iterable, Dict, Optional

from bauh.api.abstract.download import FileDownloader
from bauh.api.abstract.handler import ProcessWatcher
from bauh.api.abstract.view import MessageType
from bauh.api.http import HttpClient
from bauh.commons.html import bold
from bauh.commons.system import ProcessHandler, SimpleProcess
from bauh.gems.arch import pacman
from bauh.view.util.translation import I18n


class ArchDownloadException(Exception):
    pass


class CacheDirCreationException(ArchDownloadException):
    pass


class MultiThreadedDownloader:

    def __init__(self, file_downloader: FileDownloader, http_client: HttpClient, mirrors_available: Iterable[str],
                 mirrors_branch: str, cache_dir: str, logger: logging.Logger):
        self.downloader = file_downloader
        self.http_client = http_client
        self.mirrors = mirrors_available
        self.branch = mirrors_branch
        self.extensions = ['.tar.zst', '.tar.xz']
        self.cache_dir = cache_dir
        self.logger = logger
        self.async_downloads = []
        self.async_downloads_lock = Lock()

    def download_package_signature(self, pkg: dict, file_url: str, output_path: str, root_password: Optional[str], watcher: ProcessWatcher):
        try:
            self.logger.info("Downloading package '{}' signature".format(pkg['n']))

            sig_downloaded = self.downloader.download(file_url=file_url + '.sig', watcher=None,
                                                      output_path=output_path + '.sig',
                                                      cwd='.', root_password=root_password,
                                                      display_file_size=False,
                                                      max_threads=1)

            if not sig_downloaded:
                msg = "Could not download package '{}' signature".format(pkg['n'])
                self.logger.warning(msg)
                watcher.print('[warning] {}'.format(msg))
            else:
                msg = "Package '{}' signature successfully downloaded".format(pkg['n'])
                self.logger.info(msg)
                watcher.print(msg)
        except Exception:
            self.logger.warning("An error occurred while download package '{}' signature".format(pkg['n']))
            traceback.print_exc()

    def download_package(self, pkg: Dict[str, str], root_password: Optional[str], substatus_prefix: str, watcher: ProcessWatcher, size: int) -> bool:
        if self.mirrors and self.branch:
            pkgname = '{}-{}{}.pkg'.format(pkg['n'], pkg['v'], ('-{}'.format(pkg['a']) if pkg['a'] else ''))

            if {f for f in glob.glob(self.cache_dir + '/*') if f.split('/')[-1].startswith(pkgname)}:
                watcher.print("{} ({}) file found o cache dir {}. Skipping download.".format(pkg['n'], pkg['v'], self.cache_dir))
                return True

            arch = pkg['a'] if pkg.get('a') and pkg['a'] != 'any' else 'x86_64'

            url_base = '{}/{}/{}/{}'.format(self.branch, pkg['r'], arch, pkgname)
            base_output_path = '{}/{}'.format(self.cache_dir, pkgname)

            for mirror in self.mirrors:
                for ext in self.extensions:
                    url = '{}{}{}'.format(mirror, url_base, ext)
                    output_path = base_output_path + ext

                    watcher.print("Downloading '{}' from mirror '{}'".format(pkgname, mirror))

                    pkg_downloaded = self.downloader.download(file_url=url, watcher=watcher, output_path=output_path,
                                                              cwd='.', root_password=root_password, display_file_size=True,
                                                              substatus_prefix=substatus_prefix,
                                                              known_size=size)
                    if not pkg_downloaded:
                        watcher.print("Could not download '{}' from mirror '{}'".format(pkgname, mirror))
                    else:
                        self.logger.info("Package '{}' successfully downloaded".format(pkg['n']))
                        t = Thread(target=self.download_package_signature, args=(pkg, url, output_path, root_password, watcher), daemon=True)
                        t.start()
                        self.async_downloads_lock.acquire()
                        self.async_downloads.append(t)
                        self.async_downloads_lock.release()
                        return True
        return False

    def wait_for_async_downloads(self):
        self.async_downloads_lock.acquire()

        try:
            if self.async_downloads:
                for t in self.async_downloads:
                    t.join()

            self.async_downloads.clear()
        finally:
            self.async_downloads_lock.release()


class MultithreadedDownloadService:

    def __init__(self, file_downloader: FileDownloader, http_client: HttpClient, logger: logging.Logger, i18n: I18n):
        self.file_downloader = file_downloader
        self.http_client = http_client
        self.logger = logger
        self.i18n = i18n

    def download_packages(self, pkgs: List[str], handler: ProcessHandler, root_password: Optional[str], sizes: Dict[str, int] = None) -> int:
        ti = time.time()
        watcher = handler.watcher
        mirrors = pacman.list_available_mirrors()

        if not mirrors:
            self.logger.warning('repository mirrors seem to be not reachable')
            watcher.print('[warning] repository mirrors seem to be not reachable')
            watcher.print('[warning] multi-threaded download cancelled')
            return 0

        branch = pacman.get_mirrors_branch()

        if not branch:
            self.logger.warning('no default repository branch found')
            watcher.print('[warning] no default repository branch found')
            watcher.print('[warning] multi-threaded download cancelled')
            return 0

        cache_dir = pacman.get_cache_dir()

        if not os.path.exists(cache_dir):
            success, _ = handler.handle_simple(SimpleProcess(['mkdir', '-p', cache_dir], root_password=root_password))

            if not success:
                msg = "could not create cache dir '{}'".format(cache_dir)
                self.logger.warning(msg)
                watcher.print("[warning] {}".format(cache_dir))
                watcher.show_message(title=self.i18n['warning'].capitalize(),
                                     body=self.i18n['arch.mthread_downloaded.error.cache_dir'].format(bold(cache_dir)),
                                     type_=MessageType.WARNING)
                raise CacheDirCreationException()

        downloader = MultiThreadedDownloader(file_downloader=self.file_downloader,
                                             mirrors_available=mirrors,
                                             mirrors_branch=branch,
                                             http_client=self.http_client,
                                             logger=self.logger,
                                             cache_dir=cache_dir)

        downloaded = 0
        pkgs_data = pacman.list_download_data(pkgs)

        if not pkgs_data:
            error_msg = "Could not retrieve download data of the following packages: {}".format(', '.join(pkgs))
            watcher.print(error_msg)
            self.logger.error(error_msg)
            return 0

        for pkg in pkgs_data:
            self.logger.info('Preparing to download package: {} ({})'.format(pkg['n'], pkg['v']))
            try:
                perc = '({0:.2f}%)'.format((downloaded / (2 * len(pkgs))) * 100)
                status_prefix = '{} [{}/{}]'.format(perc, downloaded + 1, len(pkgs))

                if downloader.download_package(pkg=pkg,
                                               root_password=root_password,
                                               watcher=handler.watcher,
                                               substatus_prefix=status_prefix,
                                               size=sizes.get(pkg['n']) if sizes else None):
                    downloaded += 1
            except Exception:
                traceback.print_exc()
                watcher.show_message(title=self.i18n['error'].capitalize(),
                                     body=self.i18n['arch.mthread_downloaded.error.cancelled'],
                                     type_=MessageType.ERROR)
                raise ArchDownloadException()

        self.logger.info("Waiting for signature downloads to complete")
        downloader.wait_for_async_downloads()
        self.logger.info("Signature downloads finished")
        tf = time.time()
        self.logger.info("Download time: {0:.2f} seconds".format(tf - ti))
        return downloaded



from abc import ABC, abstractmethod
from typing import Iterable, List, Optional

from bauh.api.abstract.handler import ProcessWatcher


class FileDownloader(ABC):

    @abstractmethod
    def download(self, file_url: str, watcher: Optional[ProcessWatcher], output_path: str, cwd: str, root_password: Optional[str] = None, substatus_prefix: str = None, display_file_size: bool = True, max_threads: int = None, known_size: int = None) -> bool:
        """
        :param file_url:
        :param watcher:
        :param output_path: the downloaded file output path. Leave None for the current directory and the same file name
        :param cwd: current working directory. Leave None if does not matter.
        :param root_password: (if the output directory is protected)
        :param substatus_prefix: custom substatus prefix ('prefix downloading xpto')
        :param display_file_size: if the file size should be displayed on the substatus
        :param max_threads: maximum number of threads (only available for multi-threaded download)
        :param known_size: known file size
        :return: success / failure
        """
        pass

    @abstractmethod
    def is_multithreaded(self) -> bool:
        pass

    @abstractmethod
    def can_work(self) -> bool:
        pass

    @abstractmethod
    def get_supported_multithreaded_clients(self) -> Iterable[str]:
        pass

    @abstractmethod
    def is_multithreaded_client_available(self, name: str) -> bool:
        pass

    @abstractmethod
    def list_available_multithreaded_clients(self) -> List[str]:
        pass

    @abstractmethod
    def get_supported_clients(self) -> tuple:
        pass



from typing import Dict, Optional

from bauh.gems.debian.model import DebianPackage


def strip_maintainer_email(maintainer: str) -> str:
    return maintainer.split('<')[0].strip()


def strip_section(section: str) -> Optional[str]:
    if section:
        section_split = section.split('/')
        return section_split[1] if len(section_split) > 1 else section


def fill_show_data(pkg: DebianPackage, data: Dict[str, object]):
    if data:
        for attr, val in data.items():
            final_attr = attr.replace(' ', '_')

            if not val or val == '<none>':
                final_val = None
            else:
                if attr == 'maintainer':
                    final_val = strip_maintainer_email(str(val))
                elif attr == 'section':
                    final_attr = 'categories'
                    final_val = (strip_section(str(val)),)
                else:
                    final_val = val

            setattr(pkg, final_attr, final_val)



import os
import time
import traceback
from logging import Logger
from typing import Optional

import yaml

from bauh.gems.web import SEARCH_INDEX_FILE


class SearchIndexManager:

    def __init__(self, logger: Logger):
        self.logger = logger

    def generate(self, suggestions: dict) -> Optional[dict]:
        if suggestions:
            ti = time.time()
            index = {}

            for key, sug in suggestions.items():
                name = sug.get('name')

                if name:
                    split_name = name.lower().strip().split(' ')
                    single_name = ''.join(split_name)

                    for word in (*split_name, single_name):
                        mapped = index.get(word)

                        if not mapped:
                            mapped = set()
                            index[word] = mapped

                        mapped.add(key)

            tf = time.time()
            self.logger.info("Took {0:.4f} seconds to generate the index".format(tf - ti))

            return index

    def read(self) -> Optional[dict]:
        if os.path.exists(SEARCH_INDEX_FILE):
            with open(SEARCH_INDEX_FILE) as f:
                return yaml.safe_load(f.read())
        else:
            self.logger.warning("No search index found at {}".format(SEARCH_INDEX_FILE))

    def write(self, index: dict) -> bool:
        if index:
            self.logger.info('Preparing search index for writing')  # YAML does not work with 'sets'

            for key in index.keys():
                index[key] = list(index[key])

            try:
                self.logger.info('Writing {} indexed keys as {}'.format(len(index), SEARCH_INDEX_FILE))
                with open(SEARCH_INDEX_FILE, 'w+') as f:
                    f.write(yaml.safe_dump(index))
                self.logger.info("Search index successfully written at {}".format(SEARCH_INDEX_FILE))
                return True
            except Exception:
                self.logger.error("Could not write the search index to {}".format(SEARCH_INDEX_FILE))
                traceback.print_exc()

        return False



import argparse
from argparse import Namespace

from bauh import __app_name__, __version__


def read() -> Namespace:
    parser = argparse.ArgumentParser(prog='{}-cli'.format(__app_name__), description="CLI for Linux software management")
    parser.add_argument('-v', '--version', action='version', version='%(prog)s {}'.format(__version__))

    sub_parsers = parser.add_subparsers(dest='command', help='commands')
    updates_parser = sub_parsers.add_parser('updates', help='List available software updates')
    updates_parser.add_argument('-f', '--format', help='Command output format. Default: %(default)s', choices=['text', 'json'], default='text')

    return parser.parse_args()



import os
import re
from typing import Optional

RE_DESKTOP_EXEC = re.compile(r'(\n?\s*\w*Exec\s*=(.+))')
RE_MANY_SPACES = re.compile(r'\s+')


def find_appimage_file(folder: str) -> Optional[str]:
    for r, d, files in os.walk(folder):
        for f in files:
            if f.lower().endswith('.appimage'):
                return f'{folder}/{f}'


def replace_desktop_entry_exec_command(desktop_entry: str, appname: str, file_path: str) -> str:
    execs = RE_DESKTOP_EXEC.findall(desktop_entry)

    if not execs:
        return desktop_entry

    final_entry = desktop_entry
    treated_name = appname.strip().lower()

    for exec_groups in execs:
        full_match = exec_groups[0]

        if full_match.strip().startswith("TryExec"):  # TryExec cause issues in some DE to display the app icon
            final_entry = final_entry.replace(full_match, "")
            continue

        cmd = RE_MANY_SPACES.sub(' ', exec_groups[1].strip())
        if cmd:
            words = cmd.split(' ')
            changed = False

            for idx in range(len(words)):
                if words[idx].lower() == treated_name:
                    words[idx] = f'"{file_path}"'
                    changed = True
                    break

            if not changed:
                words = [f'"{file_path}"']

            final_entry = final_entry.replace(full_match, full_match.replace(exec_groups[1], ' '.join(words)))

    return final_entry



import traceback
from io import StringIO
from threading import Thread

from bauh.api.abstract.cache import MemoryCache
from bauh.api.abstract.context import ApplicationContext
from bauh.api.abstract.controller import SoftwareManager
from bauh.api.abstract.model import PackageStatus
from bauh.gems.flatpak.constants import FLATHUB_API_URL, FLATHUB_URL
from bauh.gems.flatpak.model import FlatpakApplication


class FlatpakAsyncDataLoader(Thread):

    def __init__(self, app: FlatpakApplication, manager: SoftwareManager, context: ApplicationContext, api_cache: MemoryCache, category_cache: MemoryCache):
        super(FlatpakAsyncDataLoader, self).__init__()
        self.app = app
        self.manager = manager
        self.http_client = context.http_client
        self.api_cache = api_cache
        self.persist = False
        self.logger = context.logger
        self.category_cache = category_cache

    @staticmethod
    def format_category(category: str) -> str:
        word = StringIO()
        last_l = None
        for idx, l in enumerate(category):
            if idx != 0 and last_l != ' ' and l.isupper() and idx + 1 < len(category) and category[idx + 1].islower():
                word.write(' ')

            last_l = l.lower()
            word.write(last_l)

        word.seek(0)
        return word.read()

    def run(self):
        if self.app:
            self.app.status = PackageStatus.LOADING_DATA

            try:
                res = self.http_client.get('{}/apps/{}'.format(FLATHUB_API_URL, self.app.id))

                if res and res.text:
                    data = res.json()

                    if not data:
                        self.logger.warning("No data returned for id {} ({})".format(self.app.id, self.app.name))
                    else:
                        if not self.app.version:
                            self.app.version = data.get('version')

                        if not self.app.name:
                            self.app.name = data.get('name')

                        if not self.app.description:
                            self.app.description = data.get('description', data.get('summary', None))

                        self.app.icon_url = data.get('iconMobileUrl', None)
                        self.app.latest_version = data.get('currentReleaseVersion', self.app.version)

                        if self.app.latest_version and (not self.app.version or not self.app.update):
                            self.app.version = self.app.latest_version

                        if not self.app.installed and self.app.latest_version:
                            self.app.version = self.app.latest_version

                        if self.app.icon_url and self.app.icon_url.startswith('/'):
                            self.app.icon_url = FLATHUB_URL + self.app.icon_url

                        if data.get('categories'):
                            cats = []
                            for c in data['categories']:
                                cached = self.category_cache.get(c['name'])

                                if not cached:
                                    cached = self.format_category(c['name'])
                                    self.category_cache.add_non_existing(c['name'], cached)

                                cats.append(cached)

                            self.app.categories = cats

                        loaded_data = self.app.get_data_to_cache()

                        self.api_cache.add(self.app.id, loaded_data)
                        self.persist = self.app.supports_disk_cache()
                else:
                    self.logger.warning("Could not retrieve app data for id '{}'. Server response: {}. Body: {}".format(self.app.id, res.status_code if res else '?', res.content.decode() if res else '?'))
            except Exception:
                self.logger.error("Could not retrieve app data for id '{}'".format(self.app.id))
                traceback.print_exc()

            self.app.status = PackageStatus.READY

            if self.persist:
                self.manager.cache_to_disk(pkg=self.app, icon_bytes=None, only_icon=False)



import re
from typing import Tuple

RE_VERSION_WITH_EPOCH = re.compile(r'^\d+:(.+)$')
RE_VERSION_WITH_RELEASE = re.compile(r'^(.+)-\d+$')


def map_str_version(version: str) -> Tuple[str, ...]:
    return tuple(part.zfill(8) for part in version.split("."))


def normalize_version(version: str) -> Tuple[int, Tuple[str, ...], int]:
    raw_version = version.strip()

    epoch = 0

    if RE_VERSION_WITH_EPOCH.match(raw_version):
        epoch_version = raw_version.split(":", maxsplit=1)
        epoch = int(epoch_version[0])
        raw_version = epoch_version[1]

    release = 1
    if RE_VERSION_WITH_RELEASE.match(raw_version):
        version_release = raw_version.rsplit("-", maxsplit=1)
        raw_version = version_release[0]
        release = int(version_release[1])

    if not raw_version.split(".")[0].isdigit():
        # this is required to properly compare versions starting with a number against versions starting with alpha
        raw_version = f"0.{raw_version}"

    return epoch, map_str_version(raw_version), release


def match_required_version(current_version: str, operator: str, required_version: str) -> bool:
    final_required, final_current = required_version.strip(), current_version.strip()

    required_has_epoch = bool(RE_VERSION_WITH_EPOCH.match(final_required))
    current_no_epoch = RE_VERSION_WITH_EPOCH.split(final_current)
    current_has_epoch = len(current_no_epoch) > 1

    if required_has_epoch and not current_has_epoch:
        final_current = f'0:{final_current}'
    elif current_has_epoch and not required_has_epoch:
        final_current = current_no_epoch[1]

    required_has_release = bool(RE_VERSION_WITH_RELEASE.match(final_required))
    current_no_release = RE_VERSION_WITH_RELEASE.split(final_current)
    current_has_release = len(current_no_release) > 1

    if required_has_release and not current_has_release:
        final_current = f'{final_current}-1'
    elif current_has_release and not required_has_release:
        final_current = current_no_release[1]

    final_required, final_current = map_str_version(final_required), map_str_version(final_current)

    if operator == '==' or operator == '=':
        return final_current == final_required
    elif operator == '>':
        return final_current > final_required
    elif operator == '>=':
        return final_current >= final_required
    elif operator == '<':
        return final_current < final_required
    elif operator == '<=':
        return final_current <= final_required



import json
import os
import re
import traceback
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta
from json import JSONDecodeError
from logging import Logger
from pathlib import Path
from typing import Optional, Set, Generator, Iterable

from bauh.commons import system
from bauh.gems.debian import APP_INDEX_FILE
from bauh.gems.debian.model import DebianApplication


class ApplicationIndexError(Exception):

    def __init__(self, cause: Optional[str] = None):
        self.cause = cause


class ApplicationIndexer:

    def __init__(self, logger: Logger, index_file_path: str = APP_INDEX_FILE):
        self._log = logger
        self._file_path = index_file_path
        self._file_timestamp_path = f'{self._file_path}.ts'

    def is_expired(self, deb_config: dict) -> bool:

        try:
            exp_minutes = int(deb_config.get('index_apps.exp', 0))
        except ValueError:
            self._log.error(f"Invalid value for Debian configuration property 'index_apps.exp': "
                            f"{deb_config['index_apps.exp']}")
            return True

        if exp_minutes <= 0:
            self._log.warning("Debian applications index will always be updated ('index_apps.exp' <= 0 )'")
            return True

        if not os.path.exists(self._file_path):
            self._log.info(f"Debian applications index not found. A new one must be generated ({self._file_path})")
            return True

        try:
            with open(self._file_timestamp_path) as f:
                timestamp_str = f.read().strip()
        except FileNotFoundError:
            self._log.info(f"Debian applications index timestamp not found ({self._file_timestamp_path})")
            return True

        try:
            timestamp = datetime.fromtimestamp(float(timestamp_str))
        except Exception:
            self._log.error(f'Could not parse the Debian applications index timestamp: {timestamp_str} '
                            f'({self._file_timestamp_path})')
            traceback.print_exc()
            return True

        expired = timestamp + timedelta(minutes=exp_minutes) <= datetime.utcnow()

        if expired:
            self._log.info("Debian applications index has expired. A new one must be generated.")
        else:
            self._log.info("Debian applications index is up-to-date")

        return expired

    def read_index(self) -> Generator[DebianApplication, None, None]:
        try:
            with open(self._file_path) as f:
                idx_str = f.read().strip()

            if not idx_str:
                self._log.warning(f"Debian applications index is empty ({self._file_path})")
            else:
                try:
                    for name, data in json.loads(idx_str).items():
                        exe_path, icon_path = data.get('exe_path'), data.get('icon_path')

                        if not all((name, exe_path, icon_path)):
                            self._log.warning(f"Invalid entry in the Debian applications index ({self._file_path}): name={name}, exe_path={exe_path}, icon_path={icon_path}")
                        else:
                            categories = data.get('categories')
                            yield DebianApplication(name=name, exe_path=exe_path, icon_path=icon_path,
                                                    categories=tuple(categories) if categories else None)

                except JSONDecodeError:
                    self._log.error(f"The Debian applications index is corrupted ({self._file_path}). "
                                    f"Could not decode the JSON.")

        except FileNotFoundError:
            self._log.warning(f"Debian applications index not found ({self._file_path})")
        except OSError as e:
            self._log.error(f"Debian applications index could not be read ({self._file_path}). OSError: {e.errno}")

    def update_index(self, apps: Set[DebianApplication], update_timestamp: bool = True):
        idx_dir = os.path.dirname(self._file_path)

        try:
            Path(idx_dir).mkdir(exist_ok=True, parents=True)
        except OSError:
            self._log.error(f"Could not create directory '{idx_dir}'")
            raise ApplicationIndexError()

        idx = {}
        if apps:
            for app in apps:
                idx.update(app.to_index())

        try:
            with open(self._file_path, 'w+') as f:
                if idx:
                    f.write(json.dumps(idx, sort_keys=True, indent=4))
                else:
                    f.write('')

        except OSError:
            self._log.error(f"Could not write to the Debian applications index file: {self._file_path}")
            raise ApplicationIndexError()

        if update_timestamp:
            index_timestamp = datetime.utcnow().timestamp()
            try:
                with open(self._file_timestamp_path, 'w+') as f:
                    f.write(str(index_timestamp))

                self._log.info("Debian applications index timestamp updated")

            except OSError:
                self._log.error(f"Could not write to the Debian applications index timestamp file: "
                                f"{self._file_timestamp_path}")
                raise ApplicationIndexError()


class ApplicationsMapper:

    def __init__(self, logger: Logger, workers: int = 10):
        self._log = logger
        self._re_desktop_file = re.compile(r'(.+):\s+(/usr/share/applications/.+\.desktop)')
        self._re_desktop_file_fields = re.compile('(Exec|TryExec|Icon|Categories|NoDisplay|Terminal)\s*=\s*(.+)')
        self._workers = workers

    def _read_file(self, file_path: str) -> Optional[str]:
        try:
            with open(file_path) as file:
                return file.read()
        except (FileNotFoundError, OSError) as e:
            self._log.error(f"Error when checking desktop file '{file_path}' ({file_path}):"
                            f" {e.__class__.__name__}")

    def _add_if_application_desktop_file(self, pkg_name: str, desktop_files: Iterable[str], output: Set[DebianApplication]):
        for file_path in sorted(desktop_files):
            content = self._read_file(file_path)

            if content:
                data = {}

                gui_app = True

                for f, v in self._re_desktop_file_fields.findall(content):
                    if f in ('NoDisplay', 'Terminal') and v.strip().lower() == 'true':
                        gui_app = False
                        break

                    if f not in data:
                        data[f.strip()] = v.strip()

                if not gui_app:
                    continue

                exe = data.get('Exec')

                if not exe:
                    exe = data.get('TryExec')

                if not exe:
                    continue

                icon = data.get('Icon')

                if not icon:
                    continue

                categories = data.get('Categories')

                if categories:
                    categories = tuple(sorted({c.strip() for c in categories.split(';') if c}))

                output.add(DebianApplication(name=pkg_name, exe_path=exe, icon_path=icon, categories=categories))
                break

    def map_executable_applications(self) -> Optional[Set[DebianApplication]]:
        exitcode, output = system.execute('dpkg-query -S .desktop', shell=True)

        if exitcode == 0 and output:

            pkg_files = dict()

            for found in self._re_desktop_file.findall(output):
                pkg_name = found[0].strip()
                files = pkg_files.get(pkg_name)

                if files is None:
                    files = set()
                    pkg_files[pkg_name] = files

                files.add(found[1].strip())

            if pkg_files:
                apps_found, check_jobs = set(), []

                with ThreadPoolExecutor(self._workers) as pool:
                    for pkg_name, files in pkg_files.items():
                        check_jobs.append(pool.submit(self._add_if_application_desktop_file,
                                                      pkg_name, files, apps_found))

                for job in check_jobs:
                    job.done()

                return apps_found



import logging
from logging import INFO

FORMAT = '%(asctime)s %(levelname)s [%(module_path)s:%(lineno)s - %(funcName)s()] - %(message)s'


class FilePathFilter(logging.Filter):

    def filter(self, record):
        record.module_path = record.pathname.split('site-packages/')[1] if 'site-packages' in record.pathname else str(record.pathname)
        return True


def new_logger(name: str, enabled: bool) -> logging.Logger:
    instance = logging.Logger(name, level=INFO)
    instance.addFilter(FilePathFilter())
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(logging.Formatter(FORMAT))
    instance.addHandler(stream_handler)
    instance.disabled = not enabled

    return instance



import logging
import os
import re
import shutil
import time
import traceback
from io import StringIO
from math import floor
from pathlib import Path
from threading import Thread
from typing import Iterable, List, Optional

from bauh.api.abstract.download import FileDownloader
from bauh.api.abstract.handler import ProcessWatcher
from bauh.api.http import HttpClient
from bauh.commons.html import bold
from bauh.commons.system import ProcessHandler, SimpleProcess
from bauh.commons.view_utils import get_human_size_str
from bauh.view.util.translation import I18n

RE_HAS_EXTENSION = re.compile(r'.+\.\w+$')


class AdaptableFileDownloader(FileDownloader):

    def __init__(self, logger: logging.Logger, multithread_enabled: bool, i18n: I18n, http_client: HttpClient,
                 multithread_client: str, check_ssl: bool):
        self.logger = logger
        self.multithread_enabled = multithread_enabled
        self.i18n = i18n
        self.http_client = http_client
        self.supported_multithread_clients = ['aria2', 'axel']
        self.multithread_client = multithread_client
        self.check_ssl = check_ssl

    @staticmethod
    def is_aria2c_available() -> bool:
        return bool(shutil.which('aria2c'))

    @staticmethod
    def is_axel_available() -> bool:
        return bool(shutil.which('axel'))

    @staticmethod
    def is_wget_available() -> bool:
        return bool(shutil.which('wget'))

    def _get_aria2c_process(self, url: str, output_path: str, cwd: str, root_password: Optional[str], threads: int) -> SimpleProcess:
        cmd = ['aria2c', url,
               '--no-conf',
               '-x', '16',
               '--enable-color=false',
               '--stderr=true',
               '--summary-interval=0',
               '--disable-ipv6',
               '-k', '1M',
               '--allow-overwrite=true',
               '-c',
               '-t', '5',
               '--max-file-not-found=3',
               '--file-allocation=none',
               '--console-log-level=error']

        if threads > 1:
            cmd.append('-s')
            cmd.append(str(threads))

        if output_path:
            output_split = output_path.split('/')
            cmd.append('-d')
            cmd.append('/'.join(output_split[:-1]))
            cmd.append('-o')
            cmd.append(output_split[-1])

        return SimpleProcess(cmd=cmd, root_password=root_password, cwd=cwd)

    def _get_axel_process(self, url: str, output_path: str, cwd: str, root_password: Optional[str], threads: int) -> SimpleProcess:
        cmd = ['axel', url, '-n', str(threads), '-4', '-c', '-T', '5']

        if not self.check_ssl:
            cmd.append('-k')

        if output_path:
            cmd.append(f'--output={output_path}')

        return SimpleProcess(cmd=cmd, cwd=cwd, root_password=root_password)

    def _get_wget_process(self, url: str, output_path: str, cwd: str, root_password: Optional[str]) -> SimpleProcess:
        cmd = ['wget', url, '-c', '--retry-connrefused', '-t', '10', '-nc']

        if not self.check_ssl:
            cmd.append('--no-check-certificate')

        if output_path:
            cmd.append('-O')
            cmd.append(output_path)

        return SimpleProcess(cmd=cmd, cwd=cwd, root_password=root_password)

    def _rm_bad_file(self, file_name: str, output_path: str, cwd, handler: ProcessHandler, root_password: Optional[str]):
        to_delete = output_path if output_path else f'{cwd}/{file_name}'

        if to_delete and os.path.exists(to_delete):
            self.logger.info(f'Removing downloaded file {to_delete}')
            success, _ = handler.handle_simple(SimpleProcess(['rm', '-rf', to_delete], root_password=root_password))
            return success

    def _concat_file_size(self, file_url: str, base_substatus: StringIO, watcher: ProcessWatcher):
        watcher.change_substatus(f'{base_substatus.getvalue()} ( ? Mb )')

        try:
            size = self.http_client.get_content_length(file_url)

            if size:
                base_substatus.write(f' ( {size} )')
                watcher.change_substatus(base_substatus.getvalue())
        except Exception:
            pass

    def _get_appropriate_threads_number(self, max_threads: int, known_size: int) -> int:
        if max_threads and max_threads > 0:
            threads = max_threads
        elif known_size:
            threads = 16 if known_size >= 16000000 else floor(known_size / 1000000)

            if threads <= 0:
                threads = 1
        else:
            threads = 16

        return threads

    def download(self, file_url: str, watcher: ProcessWatcher, output_path: str = None, cwd: str = None, root_password: Optional[str] = None, substatus_prefix: str = None, display_file_size: bool = True, max_threads: int = None, known_size: int = None) -> bool:
        self.logger.info(f'Downloading {file_url}')
        handler = ProcessHandler(watcher)
        file_name = file_url.split('/')[-1]

        final_cwd = cwd if cwd else '.'

        success = False
        ti = time.time()
        try:
            if output_path:
                if os.path.exists(output_path):
                    self.logger.info(f'Removing old file found before downloading: {output_path}')
                    os.remove(output_path)
                    self.logger.info(f'Old file {output_path} removed')
                else:
                    output_dir = os.path.dirname(output_path)

                    try:
                        Path(output_dir).mkdir(exist_ok=True, parents=True)
                    except OSError:
                        self.logger.error(f"Could not make download directory '{output_dir}'")
                        watcher.print(self.i18n['error.mkdir'].format(dir=output_dir))
                        return False

            client = self.get_available_multithreaded_tool()
            if client:
                threads = self._get_appropriate_threads_number(max_threads, known_size)

                if client == 'aria2':
                    ti = time.time()
                    process = self._get_aria2c_process(file_url, output_path, final_cwd, root_password, threads)
                    downloader = 'aria2'
                else:
                    ti = time.time()
                    process = self._get_axel_process(file_url, output_path, final_cwd, root_password, threads)
                    downloader = 'axel'
            else:
                ti = time.time()
                process = self._get_wget_process(file_url, output_path, final_cwd, root_password)
                downloader = 'wget'

            name = file_url.split('/')[-1]

            if output_path and not RE_HAS_EXTENSION.match(name) and RE_HAS_EXTENSION.match(output_path):
                name = output_path.split('/')[-1]

            if watcher:
                msg = StringIO()
                msg.write(f'{substatus_prefix} ' if substatus_prefix else '')
                msg.write(f"{bold('[{}]'.format(downloader))} {self.i18n['downloading']} {bold(name)}")

                if display_file_size:
                    if known_size:
                        msg.write(f' ( {get_human_size_str(known_size)} )')
                        watcher.change_substatus(msg.getvalue())
                    else:
                        Thread(target=self._concat_file_size, args=(file_url, msg, watcher)).start()
                else:
                    msg.write(' ( ? Mb )')
                    watcher.change_substatus(msg.getvalue())

            success, _ = handler.handle_simple(process)
        except Exception:
            traceback.print_exc()
            self._rm_bad_file(file_name, output_path, final_cwd, handler, root_password)

        tf = time.time()
        self.logger.info(f'{file_name} download took {(tf - ti) / 60:.2f} minutes')

        if not success:
            self.logger.error(f"Could not download '{file_name}'")
            self._rm_bad_file(file_name, output_path, final_cwd, handler, root_password)

        return success

    def is_multithreaded(self) -> bool:
        return bool(self.get_available_multithreaded_tool())

    def get_available_multithreaded_tool(self) -> str:
        if self.multithread_enabled:
            if self.multithread_client is None or self.multithread_client not in self.supported_multithread_clients:
                for client in self.supported_multithread_clients:
                    if self.is_multithreaded_client_available(client):
                        return client
            else:
                possible_clients = {*self.supported_multithread_clients}

                if self.is_multithreaded_client_available(self.multithread_client):
                    return self.multithread_client
                else:
                    possible_clients.remove(self.multithread_client)

                    for client in possible_clients:
                        if self.is_multithreaded_client_available(client):
                            return client

    def can_work(self) -> bool:
        return self.is_wget_available() or self.is_multithreaded()

    def get_supported_multithreaded_clients(self) -> Iterable[str]:
        return self.supported_multithread_clients

    def is_multithreaded_client_available(self, name: str) -> bool:
        if name == 'aria2':
            return self.is_aria2c_available()
        elif name == 'axel':
            return self.is_axel_available()
        else:
            return False

    def list_available_multithreaded_clients(self) -> List[str]:
        return [c for c in self.supported_multithread_clients if self.is_multithreaded_client_available(c)]

    def get_supported_clients(self) -> tuple:
        return 'wget', 'aria2', 'axel'



from bauh.commons.config import YAMLConfigManager
from bauh.gems.flatpak import CONFIG_FILE


class FlatpakConfigManager(YAMLConfigManager):

    def __init__(self):
        super(FlatpakConfigManager, self).__init__(config_file_path=CONFIG_FILE)

    def get_default_config(self) -> dict:
        return {'installation_level': None}



import logging
import time
import traceback
from typing import Optional

import requests
import yaml

from bauh.commons import system
from bauh.commons.view_utils import get_human_size_str


class HttpClient:

    def __init__(self, logger: logging.Logger, max_attempts: int = 2, timeout: int = 30, sleep: float = 0.5):
        self.max_attempts = max_attempts
        self.session = requests.Session()
        self.timeout = timeout
        self.sleep = sleep
        self.logger = logger

    def get(self, url: str, params: dict = None, headers: dict = None, allow_redirects: bool = True, ignore_ssl: bool = False, single_call: bool = False, session: bool = True) -> Optional[requests.Response]:
        cur_attempts = 1

        while cur_attempts <= self.max_attempts:
            cur_attempts += 1

            try:
                args = {'timeout': self.timeout, 'allow_redirects': allow_redirects}

                if params:
                    args['params'] = params

                if headers:
                    args['headers'] = headers

                if ignore_ssl:
                    args['verify'] = False

                if session:
                    res = self.session.get(url, **args)
                else:
                    res = requests.get(url, **args)

                if res.status_code == 200:
                    return res

                if single_call:
                    return res

                if self.sleep > 0:
                    time.sleep(self.sleep)
            except Exception as e:
                if isinstance(e, requests.exceptions.ConnectionError):
                    self.logger.error('Internet seems to be off')
                    raise e
                elif isinstance(e, requests.exceptions.TooManyRedirects):
                    self.logger.warning(f"Too many redirects for GET -> {url}")
                    raise e
                elif e.__class__ in (requests.exceptions.MissingSchema, requests.exceptions.InvalidSchema):
                    self.logger.warning(f"The URL '{url}' has an invalid schema")
                    raise e

                self.logger.error(f"Could not retrieve data from '{url}'")
                traceback.print_exc()
                continue

            self.logger.warning(f"Could not retrieve data from '{url}'")

    def get_json(self, url: str, params: dict = None, headers: dict = None, allow_redirects: bool = True, session: bool = True):
        res = self.get(url, params=params, headers=headers, allow_redirects=allow_redirects, session=session)
        return res.json() if res else None

    def get_yaml(self, url: str, params: dict = None, headers: dict = None, allow_redirects: bool = True, session: bool = True):
        res = self.get(url, params=params, headers=headers, allow_redirects=allow_redirects, session=session)
        return yaml.safe_load(res.text) if res else None

    def get_content_length_in_bytes(self, url: str, session: bool = True) -> Optional[int]:
        if not url:
            return

        params = {'url': url, 'allow_redirects': True, 'stream': True}

        try:
            if session:
                res = self.session.get(**params)
            else:
                res = requests.get(**params)
        except requests.exceptions.ConnectionError:
            self.logger.info(f"Internet seems to be off. Could not reach '{url}'")
            return

        if res.status_code == 200:
            size = res.headers.get('Content-Length')

            if size:
                try:
                    return int(size)
                except Exception:
                    pass

    def get_content_length(self, url: str, session: bool = True) -> Optional[str]:
        size = self.get_content_length_in_bytes(url, session)

        if size:
            return get_human_size_str(size)

    def exists(self, url: str, session: bool = True, timeout: int = 5) -> bool:
        params = {'url': url, 'allow_redirects': True, 'verify': False, 'timeout': timeout}

        try:
            if session:
                res = self.session.head(**params)
            else:
                res = self.session.get(**params)
        except requests.exceptions.TooManyRedirects:
            self.logger.warning(f"{url} seems to exist, but too many redirects have happened")
            return True

        return res.status_code in (200, 403)



import logging
import os
import time
import traceback
from datetime import datetime
from logging import Logger
from pathlib import Path

from bauh.api.paths import CACHE_DIR

SYNC_FILE = f'{CACHE_DIR}/arch/mirrors_sync'


def should_sync(logger: logging.Logger):
    if os.path.exists(SYNC_FILE):
        with open(SYNC_FILE) as f:
            sync_file = f.read()

        try:
            sync_time = datetime.fromtimestamp(int(sync_file))
            now = datetime.now()

            if now > sync_time and now.day != sync_time.day:
                logger.info("Package databases synchronization out of date")
            else:
                msg = "Package databases already synchronized"
                logger.info(msg)
                return False
        except Exception:
            logger.warning("Could not convert the database synchronization time from '{}".format(SYNC_FILE))
            traceback.print_exc()
    return True


def register_sync(logger: Logger):
    try:
        Path('/'.join(SYNC_FILE.split('/')[0:-1])).mkdir(parents=True, exist_ok=True)
        with open(SYNC_FILE, 'w+') as f:
            f.write(str(int(time.time())))
    except Exception:
        logger.error("Could not write to mirrors sync file '{}'".format(SYNC_FILE))
        traceback.print_exc()



from __future__ import annotations
import glob
import logging
import os
import re
import shutil
import time
import traceback
from datetime import datetime, timedelta
from pathlib import Path
from threading import Thread
from typing import Optional

import requests

from bauh.api.abstract.context import ApplicationContext
from bauh.api.abstract.handler import TaskManager
from bauh.commons.boot import CreateConfigFile
from bauh.commons.html import bold
from bauh.commons.system import new_root_subprocess, ProcessHandler
from bauh.gems.arch import pacman, disk, CUSTOM_MAKEPKG_FILE, ARCH_CONFIG_DIR, AUR_INDEX_FILE, get_icon_path, database, \
    mirrors, ARCH_CACHE_DIR, AUR_INDEX_TS_FILE, aur
from bauh.gems.arch.aur import URL_INDEX
from bauh.view.util.translation import I18n

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from bauh.gems.arch.controller import ArchManager

URL_INFO = 'https://aur.archlinux.org/rpc/?v=5&type=info&arg={}'

GLOBAL_MAKEPKG = '/etc/makepkg.conf'

RE_MAKE_FLAGS = re.compile(r'#?\s*MAKEFLAGS\s*=\s*.+\s*')
RE_CLEAR_REPLACE = re.compile(r'[\-_.]')


class AURIndexUpdater(Thread):

    def __init__(self, context: ApplicationContext, taskman: TaskManager, create_config: Optional[CreateConfigFile] = None, arch_config: Optional[dict] = None):
        super(AURIndexUpdater, self).__init__(daemon=True)
        self.http_client = context.http_client
        self.i18n = context.i18n
        self.logger = context.logger
        self.taskman = taskman
        self.task_id = 'index_aur'
        self.create_config = create_config
        self.config = arch_config
        self.taskman.register_task(self.task_id, self.i18n['arch.task.aur.index.status'], get_icon_path())

    def should_update(self) -> bool:
        if not aur.is_supported(self.config):
            return False

        try:
            exp_hours = int(self.config['aur_idx_exp'])
        except Exception:
            traceback.print_exc()
            return True

        if exp_hours <= 0:
            return True

        if not os.path.exists(AUR_INDEX_FILE):
            return True

        if not os.path.exists(AUR_INDEX_TS_FILE):
            return True

        with open(AUR_INDEX_TS_FILE) as f:
            timestamp_str = f.read()

        try:
            index_timestamp = datetime.fromtimestamp(float(timestamp_str))
            return (index_timestamp + timedelta(hours=exp_hours)) <= datetime.utcnow()
        except Exception:
            traceback.print_exc()
            return True

    def update_index(self):
        self.logger.info('Indexing AUR packages')
        self.taskman.update_progress(self.task_id, 5, self.i18n['arch.task.aur.index.substatus.download'])
        try:
            index_ts = datetime.utcnow().timestamp()
            res = self.http_client.get(URL_INDEX)

            if res and res.text:
                index_progress = 50
                self.taskman.update_progress(self.task_id, index_progress,
                                             self.i18n['arch.task.aur.index.substatus.gen_index'])
                indexed = 0

                Path(os.path.dirname(AUR_INDEX_FILE)).mkdir(parents=True, exist_ok=True)

                with open(AUR_INDEX_FILE, 'w+') as f:
                    lines = res.text.split('\n')
                    progress_inc = round(len(lines) / 50)  # 1%

                    perc_count = 0
                    for n in lines:
                        if index_progress < 100 and perc_count == progress_inc:
                            index_progress += 1
                            perc_count = 0
                            self.taskman.update_progress(self.task_id, index_progress,
                                                         self.i18n['arch.task.aur.index.substatus.gen_index'])

                        if n and not n.startswith('#'):
                            f.write('{}={}\n'.format(RE_CLEAR_REPLACE.sub('', n), n))
                            indexed += 1

                        perc_count += 1

                with open(AUR_INDEX_TS_FILE, 'w+') as f:
                    f.write(str(index_ts))

                self.logger.info('Pre-indexed {} AUR package names at {}'.format(indexed, AUR_INDEX_FILE))
                self.taskman.update_progress(self.task_id, 100, None)

            else:
                self.logger.warning('No data returned from: {}'.format(URL_INDEX))
                self.taskman.update_progress(self.task_id, 100, self.i18n['arch.task.aur.index.substatus.error.no_data'])

        except requests.exceptions.ConnectionError:
            self.logger.warning('No internet connection: could not pre-index packages')
            self.taskman.update_progress(self.task_id, 100, self.i18n['arch.task.aur.index.substatus.error.download'])

    def run(self):
        ti = time.time()

        if self.create_config:
            self.taskman.update_progress(self.task_id, 0, self.i18n['task.waiting_task'].format(self.create_config.task_name))
            self.create_config.join()
            self.config = self.create_config.config

        self.taskman.update_progress(self.task_id, 1, self.i18n['arch.task.aur.index.substatus.checking'])

        if self.should_update():
            self.update_index()
        else:
            self.logger.info("AUR index is up to date. Aborting...")
            self.taskman.update_progress(self.task_id, 100, None)

        tf = time.time()
        self.taskman.finish_task(self.task_id)
        self.logger.info("Finished. Took {0:.5f} seconds".format(tf - ti))


class ArchDiskCacheUpdater(Thread):

    def __init__(self, taskman: TaskManager, i18n: I18n, logger: logging.Logger,
                 controller: ArchManager, internet_available: bool, aur_indexer: Thread,
                 create_config: CreateConfigFile):
        super(ArchDiskCacheUpdater, self).__init__(daemon=True)
        self.logger = logger
        self.taskman = taskman
        self.task_id = 'arch_cache_up'
        self.i18n = i18n
        self.indexed = 0
        self.indexed_template = self.i18n['arch.task.disk_cache.indexed'] + ': {}/ {}'
        self.to_index = 0
        self.progress = 0  # progress is defined by the number of packages prepared and indexed
        self.controller = controller
        self.internet_available = internet_available
        self.installed_hash_path = f'{ARCH_CACHE_DIR}/installed.sha1'
        self.installed_cache_dir = f'{ARCH_CACHE_DIR}/installed'
        self.aur_indexer = aur_indexer
        self.create_config = create_config
        self.taskman.register_task(self.task_id, self.i18n['arch.task.disk_cache'], get_icon_path())

    def update_indexed(self, pkgname: str):
        self.indexed += 1
        sub = self.indexed_template.format(self.indexed, self.to_index)
        progress = self.progress + (self.indexed / self.to_index) * 50
        self.taskman.update_progress(self.task_id, progress, sub)

    def _update_progress(self, progress: float, msg: str):
        self.progress = progress
        self.taskman.update_progress(self.task_id, self.progress, msg)

    def _notify_reading_files(self):
        self._update_progress(50, self.i18n['arch.task.disk_cache.indexing'])

    def run(self):
        ti = time.time()
        self.taskman.update_progress(self.task_id, 0, self.i18n['task.waiting_task'].format(self.create_config.task_name))
        self.create_config.join()

        config = self.create_config.config
        aur_supported, repositories = aur.is_supported(config), config['repositories']

        self.taskman.update_progress(self.task_id, 1, None)
        if not any([aur_supported, repositories]):
            self.taskman.update_progress(self.task_id, 100, self.i18n['arch.task.disabled'])
            self.taskman.finish_task(self.task_id)
            return

        self.logger.info("Checking already cached package data")
        self._update_progress(1, self.i18n['arch.task.disk_cache.checking'])

        cache_dirs = [fpath for fpath in glob.glob('{}/*'.format(self.installed_cache_dir)) if os.path.isdir(fpath)]

        not_cached_names = None

        self._update_progress(15, self.i18n['arch.task.disk_cache.checking'])
        if cache_dirs:  # if there are cache data
            installed_names = pacman.list_installed_names()
            cached_pkgs = {cache_dir.split('/')[-1] for cache_dir in cache_dirs}

            not_cached_names = installed_names.difference(cached_pkgs)
            self._update_progress(20, self.i18n['arch.task.disk_cache.checking'])

            if not not_cached_names:
                self.taskman.update_progress(self.task_id, 100, '')
                self.taskman.finish_task(self.task_id)
                tf = time.time()
                time_msg = '{0:.2f} seconds'.format(tf - ti)
                self.logger.info('Finished: no package data to cache ({})'.format(time_msg))
                return

        self.logger.info('Pre-caching installed Arch packages data to disk')

        if aur_supported and self.aur_indexer:
            self.taskman.update_progress(self.task_id, 20, self.i18n['arch.task.disk_cache.waiting_aur_index'].format(bold(self.i18n['arch.task.aur.index.status'])))
            self.aur_indexer.join()

        self._update_progress(21, self.i18n['arch.task.disk_cache.checking'])
        installed = self.controller.read_installed(disk_loader=None, internet_available=self.internet_available,
                                                   only_apps=False, pkg_types=None, limit=-1, names=not_cached_names,
                                                   wait_disk_cache=False).installed

        self._update_progress(35, self.i18n['arch.task.disk_cache.checking'])

        saved = 0

        pkgs = {p.name: p for p in installed if ((aur_supported and p.repository == 'aur') or (repositories and p.repository != 'aur')) and not os.path.exists(p.get_disk_cache_path())}
        self.to_index = len(pkgs)

        # overwrite == True because the verification already happened
        self._update_progress(40, self.i18n['arch.task.disk_cache.reading_files'])
        saved += disk.write_several(pkgs=pkgs,
                                    after_desktop_files=self._notify_reading_files,
                                    after_written=self.update_indexed, overwrite=True)
        self.taskman.update_progress(self.task_id, 100, None)
        self.taskman.finish_task(self.task_id)

        tf = time.time()
        time_msg = '{0:.2f} seconds'.format(tf - ti)
        self.logger.info('Finished: pre-cached data of {} Arch packages to the disk ({})'.format(saved, time_msg))


class ArchCompilationOptimizer(Thread):

    def __init__(self, i18n: I18n, logger: logging.Logger, taskman: TaskManager, create_config: Optional[CreateConfigFile] = None):
        super(ArchCompilationOptimizer, self).__init__(daemon=True)
        self.logger = logger
        self.i18n = i18n
        self.re_compress_xz = re.compile(r'#?\s*COMPRESSXZ\s*=\s*.+')
        self.re_compress_zst = re.compile(r'#?\s*COMPRESSZST\s*=\s*.+')
        self.re_build_env = re.compile(r'\s+BUILDENV\s*=.+')
        self.re_ccache = re.compile(r'!?ccache')
        self.taskman = taskman
        self.task_id = 'arch_make_optm'
        self.create_config = create_config
        self.taskman.register_task(self.task_id, self.i18n['arch.task.optimizing'].format(bold('makepkg.conf')), get_icon_path())

    def _is_ccache_installed(self) -> bool:
        return bool(shutil.which('ccache'))

    def optimize(self):
        ti = time.time()
        try:
            ncpus = os.cpu_count()
        except Exception:
            self.logger.error('Could not determine the number of processors. Aborting...')
            ncpus = None

        if os.path.exists(GLOBAL_MAKEPKG):
            self.logger.info("Verifying if it is possible to optimize Arch packages compilation")

            with open(GLOBAL_MAKEPKG) as f:
                global_makepkg = f.read()

            Path(ARCH_CONFIG_DIR).mkdir(parents=True, exist_ok=True)

            custom_makepkg, optimizations = None, []

            if ncpus:
                makeflags = RE_MAKE_FLAGS.findall(global_makepkg)

                if makeflags:
                    not_commented = [f for f in makeflags if not f.startswith('#')]

                    if not not_commented:
                        custom_makepkg = RE_MAKE_FLAGS.sub('', global_makepkg)
                        optimizations.append('MAKEFLAGS="-j$(nproc)"')
                    else:
                        self.logger.warning("It seems '{}' compilation flags are already customized".format(GLOBAL_MAKEPKG))
                else:
                    optimizations.append('MAKEFLAGS="-j$(nproc)"')

            self.taskman.update_progress(self.task_id, 20, None)

            compress_xz = self.re_compress_xz.findall(custom_makepkg or global_makepkg)

            if compress_xz:
                not_eligible = [f for f in compress_xz if not f.startswith('#') and '--threads' in f]

                if not not_eligible:
                    custom_makepkg = self.re_compress_xz.sub('', custom_makepkg or global_makepkg)
                    optimizations.append('COMPRESSXZ=(xz -c -z - --threads=0)')
                else:
                    self.logger.warning("It seems '{}' COMPRESSXZ is already customized".format(GLOBAL_MAKEPKG))
            else:
                optimizations.append('COMPRESSXZ=(xz -c -z - --threads=0)')

            self.taskman.update_progress(self.task_id, 40, None)

            compress_zst = self.re_compress_zst.findall(custom_makepkg or global_makepkg)

            if compress_zst:
                not_eligible = [f for f in compress_zst if not f.startswith('#') and '--threads' in f]

                if not not_eligible:
                    custom_makepkg = self.re_compress_zst.sub('', custom_makepkg or global_makepkg)
                    optimizations.append('COMPRESSZST=(zstd -c -z -q - --threads=0)')
                else:
                    self.logger.warning("It seems '{}' COMPRESSZST is already customized".format(GLOBAL_MAKEPKG))
            else:
                optimizations.append('COMPRESSZST=(zstd -c -z -q - --threads=0)')

            self.taskman.update_progress(self.task_id, 60, None)

            build_envs = self.re_build_env.findall(custom_makepkg or global_makepkg)

            if build_envs:
                build_def = None
                for e in build_envs:
                    env_line = e.strip()

                    ccache_defs = self.re_ccache.findall(env_line)
                    ccache_installed = self._is_ccache_installed()

                    if ccache_defs:
                        if ccache_installed:
                            custom_makepkg = (custom_makepkg or global_makepkg).replace(e, '')

                            if not build_def:
                                build_def = self.re_ccache.sub('', env_line).replace('(', '(ccache ')
                        elif not build_def:
                            build_def = self.re_ccache.sub('', env_line)

                if build_def:
                    optimizations.append(build_def)
            else:
                self.logger.warning("No BUILDENV declaration found")

                if self._is_ccache_installed():
                    self.logger.info('Adding a BUILDENV declaration')
                    optimizations.append('BUILDENV=(ccache)')

            self.taskman.update_progress(self.task_id, 80, None)

            if custom_makepkg and optimizations:
                generated_by = '# <generated by bauh>\n'
                custom_makepkg = custom_makepkg + '\n' + generated_by + '\n'.join(optimizations) + '\n'

                with open(CUSTOM_MAKEPKG_FILE, 'w+') as f:
                    f.write(custom_makepkg)

                self.logger.info("A custom optimized 'makepkg.conf' was generated at '{}'".format(CUSTOM_MAKEPKG_FILE))
            else:
                self.logger.info("No optimizations are necessary")

                if os.path.exists(CUSTOM_MAKEPKG_FILE):
                    self.logger.info("Removing old optimized 'makepkg.conf' at '{}'".format(CUSTOM_MAKEPKG_FILE))
                    os.remove(CUSTOM_MAKEPKG_FILE)

            self.taskman.update_progress(self.task_id, 100, None)
            tf = time.time()
            self.logger.info('Finished. {0:.2f} seconds'.format(tf - ti))

    def run(self):
        ti = time.time()
        if self.create_config:
            self.taskman.update_progress(self.task_id, 0, self.i18n['task.waiting_task'].format(bold(self.create_config.task_name)))
            self.create_config.join()

            self.taskman.update_progress(self.task_id, 1, None)

            if self.create_config.config['optimize'] and aur.is_supported(self.create_config.config):
                try:
                    self.optimize()
                except Exception:
                    self.logger.error("Unexpected exception")
                    traceback.print_exc()
                    self.taskman.update_progress(self.task_id, 100, None)
            else:
                self.logger.info("AUR packages compilation optimizations are disabled")

                if os.path.exists(CUSTOM_MAKEPKG_FILE):
                    try:
                        self.logger.info("Removing custom 'makepkg.conf' -> '{}'".format(CUSTOM_MAKEPKG_FILE))
                        os.remove(CUSTOM_MAKEPKG_FILE)
                    except Exception:
                        self.logger.error("Unexpected exception")
                        traceback.print_exc()

                self.taskman.update_progress(self.task_id, 100, self.i18n['arch.task.disabled'])

        tf = time.time()
        self.taskman.finish_task(self.task_id)
        self.logger.info('Finished. Took {0:.2f} seconds'.format(tf - ti))


class RefreshMirrors(Thread):

    def __init__(self, taskman: TaskManager, root_password: Optional[str], i18n: I18n, logger: logging.Logger,
                 create_config: CreateConfigFile):
        super(RefreshMirrors, self).__init__(daemon=True)
        self.taskman = taskman
        self.i18n = i18n
        self.logger = logger
        self.root_password = root_password
        self.task_id = "arch_mirrors"
        self.create_config = create_config
        self.refreshed = False
        self.task_name = self.i18n['arch.task.mirrors']
        self.taskman.register_task(self.task_id, self.task_name, get_icon_path())

    def _notify_output(self, output: str):
        self.taskman.update_output(self.task_id, output)

    @staticmethod
    def is_enabled(arch_config: dict, aur_supported: bool) -> bool:
        return (arch_config['repositories'] or aur_supported) \
            and arch_config['refresh_mirrors_startup'] and pacman.is_mirrors_available()

    @classmethod
    def should_synchronize(cls, arch_config: dict, aur_supported: bool, logger: logging.Logger) -> bool:
        return cls.is_enabled(arch_config, aur_supported) and mirrors.should_sync(logger)

    def run(self):
        ti = time.time()
        self.taskman.update_progress(self.task_id, 0, self.i18n['task.waiting_task'].format(bold(self.create_config.task_name)))
        self.create_config.join()

        arch_config = self.create_config.config
        aur_supported = aur.is_supported(arch_config)

        self.taskman.update_progress(self.task_id, 1, self.i18n['arch.task.checking_settings'])

        if not self.is_enabled(arch_config, aur_supported):
            self.taskman.update_progress(self.task_id, 100, self.i18n['arch.task.disabled'])
            self.taskman.finish_task(self.task_id)
            return

        if not mirrors.should_sync(self.logger):
            self.taskman.update_progress(self.task_id, 100, self.i18n['arch.task.mirrors.cached'])
            self.taskman.finish_task(self.task_id)
            return

        sort_limit = arch_config['mirrors_sort_limit']
        self.logger.info("Refreshing mirrors")

        handler = ProcessHandler()
        try:
            self.taskman.update_progress(self.task_id, 10, '')
            success, output = handler.handle_simple(pacman.refresh_mirrors(self.root_password), output_handler=self._notify_output)

            if success:

                if sort_limit is not None and sort_limit >= 0:
                    self.taskman.update_progress(self.task_id, 50, self.i18n['arch.custom_action.refresh_mirrors.status.updating'])
                    try:
                        handler.handle_simple(pacman.sort_fastest_mirrors(self.root_password, sort_limit), output_handler=self._notify_output)
                    except Exception:
                        self.logger.error("Could not sort mirrors by speed")
                        traceback.print_exc()

                mirrors.register_sync(self.logger)
                self.refreshed = True
            else:
                self.logger.error("It was not possible to refresh mirrors")
        except Exception:
            self.logger.error("It was not possible to refresh mirrors")
            traceback.print_exc()

        self.taskman.update_progress(self.task_id, 100, None)
        self.taskman.finish_task(self.task_id)
        tf = time.time()
        self.logger.info("Finished. Took {0:.2f} seconds".format(tf - ti))


class SyncDatabases(Thread):

    def __init__(self, taskman: TaskManager, root_password: Optional[str], i18n: I18n, logger: logging.Logger,
                 refresh_mirrors: RefreshMirrors, create_config: CreateConfigFile):
        super(SyncDatabases, self).__init__(daemon=True)
        self.task_man = taskman
        self.i18n = i18n
        self.taskman = taskman
        self.task_id = "arch_dbsync"
        self.root_password = root_password
        self.refresh_mirrors = refresh_mirrors
        self.logger = logger
        self.create_config = create_config
        self.task_name = self.i18n['arch.sync_databases.substatus']
        self.taskman.register_task(self.task_id, self.task_name, get_icon_path())
        self.synchronized = False

    @staticmethod
    def is_enabled(arch_config: dict, aur_supported: bool) -> bool:
        return arch_config['sync_databases_startup'] and (aur_supported or arch_config['repositories'])

    @classmethod
    def should_sync(cls, mirrors_refreshed: bool, arch_config: dict, aur_supported: bool, logger: logging.Logger):
        return mirrors_refreshed or (cls.is_enabled(arch_config, aur_supported) and database.should_sync(arch_config, aur_supported, None, logger))

    def run(self) -> None:
        self.taskman.update_progress(self.task_id, 0, self.i18n['task.waiting_task'].format(bold(self.create_config.task_name)))
        self.create_config.join()

        self.taskman.update_progress(self.task_id, 0, self.i18n['task.waiting_task'].format(bold(self.refresh_mirrors.task_name)))
        self.refresh_mirrors.join()

        self.taskman.update_progress(self.task_id, 1, self.i18n['arch.task.checking_settings'])

        arch_config = self.create_config.config
        aur_supported = aur.is_supported(arch_config)
        if not self.is_enabled(arch_config, aur_supported):
            self.taskman.update_progress(self.task_id, 100, self.i18n['arch.task.disabled'])
            self.taskman.finish_task(self.task_id)
            return

        shoud_sync = self.refresh_mirrors.refreshed or (database.should_sync(arch_config, aur_supported, None, self.logger))

        if not shoud_sync:
            self.taskman.update_progress(self.task_id, 100, self.i18n['arch.sync_databases.substatus.synchronized'])
            self.taskman.finish_task(self.task_id)
            self.synchronized = True
            return

        self.logger.info("Synchronizing databases")
        self.taskman.register_task(self.task_id, self.i18n['arch.sync_databases.substatus'], get_icon_path())

        progress = 10
        dbs = pacman.get_databases()
        self.taskman.update_progress(self.task_id, progress, None)

        if dbs:
            inc = 90 / len(dbs)
            try:
                p = new_root_subprocess(['pacman', '-Syy'], self.root_password)

                dbs_read, last_db = 0, None

                for o in p.stdout:
                    line = o.decode().strip()

                    if line:
                        self.task_man.update_output(self.task_id, line)
                        if line.startswith('downloading'):
                            db = line.split(' ')[1].strip()

                            if last_db is None or last_db != db:
                                last_db = db
                                dbs_read += 1
                                progress = dbs_read * inc
                            else:
                                progress += 0.25

                            self.taskman.update_progress(self.task_id, progress, self.i18n['arch.task.sync_sb.status'].format(db))

                for o in p.stderr:
                    line = o.decode().strip()

                    if line:
                        self.task_man.update_output(self.task_id, line)

                p.wait()

                if p.returncode == 0:
                    database.register_sync(self.logger)
                    self.synchronized = True
                else:
                    self.logger.error("Could not synchronize database")

            except Exception:
                self.logger.info("Error while synchronizing databases")
                traceback.print_exc()

        self.taskman.update_progress(self.task_id, 100, None)
        self.taskman.finish_task(self.task_id)
        self.logger.info("Finished")



import os
from pathlib import Path

from bauh.api import user
from bauh.api.paths import CONFIG_DIR
from bauh.commons import resource
from bauh.commons.version_util import map_str_version

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
CONFIG_FILE = f'{CONFIG_DIR}/flatpak.yml'
FLATPAK_CONFIG_DIR = f'{CONFIG_DIR}/flatpak'
UPDATES_IGNORED_FILE = f'{FLATPAK_CONFIG_DIR}/updates_ignored.txt'
EXPORTS_PATH = '/usr/share/flatpak/exports/share' if user.is_root() else f'{Path.home()}/.local/share/flatpak/exports/share'
VERSION_1_2 = map_str_version("1.2")
VERSION_1_3 = map_str_version("1.3")
VERSION_1_4 = map_str_version("1.4")
VERSION_1_5 = map_str_version("1.5")
VERSION_1_12 = map_str_version("1.12")


def get_icon_path() -> str:
    return resource.get_path('img/flatpak.svg', ROOT_DIR)



import shutil
from io import StringIO
from logging import Logger
from typing import List, Tuple, Optional

from bauh.commons import system
from bauh.commons.system import SimpleProcess


def is_installed() -> bool:
    return bool(shutil.which('git'))


def list_commits(proj_dir: str, limit: int = -1, logger: Optional[Logger] = None) -> Optional[List[Tuple[str, int]]]:
    if limit == 0:
        return

    cmd = StringIO()
    cmd.write('git log --format="%H %ct"')

    if limit > 0:
        cmd.write(f' -{limit}')

    code, output = system.execute(cmd.getvalue(), cwd=proj_dir, shell=True)

    if code == 0 and output:
        commits = []
        for line in output.split('\n'):
            line_strip = line.strip()

            if line_strip:
                line_split = line_strip.split(' ', 1)

                if len(line_split) == 2:
                    commit_sha = line_split[0].strip()
                    try:
                        commit_date = int(line_split[1].strip())
                    except ValueError:
                        commit_date = None

                        if logger:
                            logger.error(f"Could not parse commit date {line_split[1]}")

                    commits.append((commit_sha, commit_date))

        return commits


def clone(url: str, target_dir: Optional[str], depth: int = -1, custom_user: Optional[str] = None) -> SimpleProcess:
    cmd = ['git', 'clone', url]

    if depth > 0:
        cmd.append(f'--depth={depth}')

    if target_dir:
        cmd.append(target_dir)

    return SimpleProcess(cmd=cmd, custom_user=custom_user)



import logging
import time
import traceback
from datetime import datetime
from threading import Thread
from typing import Optional

from bauh.api.abstract.handler import TaskManager
from bauh.commons.boot import CreateConfigFile
from bauh.commons.html import bold
from bauh.gems.web import get_icon_path
from bauh.gems.web.environment import EnvironmentUpdater
from bauh.gems.web.search import SearchIndexManager
from bauh.gems.web.suggestions import SuggestionsManager
from bauh.view.util.translation import I18n


class SuggestionsLoader(Thread):

    def __init__(self, taskman: TaskManager, manager: SuggestionsManager,
                 i18n: I18n, logger: logging.Logger, suggestions_callback, create_config: CreateConfigFile,
                 internet_connection: bool, suggestions: Optional[dict] = None):
        super(SuggestionsLoader, self).__init__(daemon=True)
        self.taskman = taskman
        self.task_id = 'web_sugs'
        self.manager = manager
        self.suggestions_callback = suggestions_callback
        self.i18n = i18n
        self.logger = logger
        self.suggestions = suggestions
        self.create_config = create_config
        self.internet_connection = internet_connection
        self.task_name = self.i18n['task.download_suggestions']
        self.taskman.register_task(self.task_id, self.task_name, get_icon_path())

    def run(self):
        ti = time.time()
        if self.manager.is_custom_local_file_mapped():
            self.taskman.update_progress(self.task_id, 50, None)
            self.logger.info(f"Local Web suggestions file mapped: {self.manager.file_url}")
            self.suggestions = self.manager.read_cached()
        else:
            wait_msg = self.i18n['task.waiting_task'].format(bold(self.create_config.task_name))
            self.taskman.update_progress(self.task_id, 0, wait_msg)
            self.create_config.join()

            self.taskman.update_progress(self.task_id, 10, None)

            if not self.internet_connection:
                self.logger.warning("No internet connection. Only cached suggestions can be loaded")
                self.suggestions = self.manager.read_cached()
            elif not self.manager.should_download(self.create_config.config):
                self.suggestions = self.manager.read_cached(check_file=False)
            else:
                try:
                    timestamp = datetime.utcnow().timestamp()
                    self.suggestions = self.manager.download()

                    if self.suggestions:
                        self.taskman.update_progress(self.task_id, 50, self.i18n['web.task.suggestions.saving'])
                        self.manager.save_to_disk(self.suggestions, timestamp)
                except Exception:
                    self.logger.error("Unexpected exception")
                    traceback.print_exc()

        if self.suggestions_callback:
            self.taskman.update_progress(self.task_id, 75, None)
            try:
                self.suggestions_callback(self.suggestions)
            except Exception:
                self.logger.error("Unexpected exception")
                traceback.print_exc()

        self.taskman.update_progress(self.task_id, 100, None)
        self.taskman.finish_task(self.task_id)
        tf = time.time()
        self.logger.info("Finished. Took {0:.4f} seconds".format(tf - ti))


class SearchIndexGenerator(Thread):

    def __init__(self, taskman: TaskManager, idxman: SearchIndexManager, suggestions_loader: SuggestionsLoader,
                 i18n: I18n, logger: logging.Logger):
        super(SearchIndexGenerator, self).__init__(daemon=True)
        self.taskman = taskman
        self.idxman = idxman
        self.i18n = i18n
        self.logger = logger
        self.suggestions_loader = suggestions_loader
        self.task_id = 'web_idx_gen'
        self.taskman.register_task(self.task_id, self.i18n['web.task.search_index'], get_icon_path())

    def run(self):
        ti = time.time()

        if self.suggestions_loader.is_alive():
            wait_msg = self.i18n['task.waiting_task'].format(bold(self.suggestions_loader.task_name))
            self.taskman.update_progress(self.task_id, 0, wait_msg)
            self.suggestions_loader.join()

        if self.suggestions_loader.suggestions:
            self.taskman.update_progress(self.task_id, 1, None)
            self.logger.info('Indexing suggestions')
            index = self.idxman.generate(self.suggestions_loader.suggestions)

            if index:
                self.taskman.update_progress(self.task_id, 50, self.i18n['web.task.suggestions.saving'])
                self.idxman.write(index)

        self.taskman.update_progress(self.task_id, 100, None)
        self.taskman.finish_task(self.task_id)
        tf = time.time()
        self.logger.info("Finished. Took {0:.4f} seconds".format(tf - ti))


class UpdateEnvironmentSettings(Thread):

    def __init__(self, env_updater: EnvironmentUpdater, taskman: TaskManager, i18n: I18n, create_config: CreateConfigFile):
        super(UpdateEnvironmentSettings, self).__init__(daemon=True)
        self.env_updater = env_updater
        self.taskman = taskman
        self.create_config = create_config
        self.task_id = env_updater.task_read_settings_id
        self.i18n = i18n

    def run(self):
        self.taskman.register_task(self.task_id, self.i18n['web.task.download_settings'], get_icon_path())
        self.taskman.update_progress(self.task_id, 1, self.i18n['task.waiting_task'].format(bold(self.create_config.task_name)))
        self.create_config.join()

        web_config = self.create_config.config
        if self.env_updater.should_download_settings(web_config):
            self.env_updater.read_settings(web_config=web_config, cache=False)
        else:
            self.taskman.update_progress(self.task_id, 100, None)
            self.taskman.finish_task(self.task_id)



import logging
import os
from pathlib import Path

from bauh import __app_name__, __version__
from bauh.api.http import HttpClient
from bauh.api.paths import CACHE_DIR
from bauh.commons.html import bold, link
from bauh.commons.version_util import normalize_version
from bauh.view.util.translation import I18n


def check_for_update(logger: logging.Logger, http_client: HttpClient, i18n: I18n, tray: bool = False) -> str:
    """
    :param logger:
    :param http_client:
    :param i18n:
    :param file_prefix: notification file prefix
    :return: bauh update warning string or 'None' if no update is available
    """
    logger.info("Checking for updates")

    try:
        releases = http_client.get_json('https://api.github.com/repos/vinifmor/bauh/releases')

        if releases:
            latest = None

            for r in releases:
                if not r['draft']:
                    latest = r
                    break

            if latest and latest.get('tag_name'):
                notifications_dir = f'{CACHE_DIR}/updates'
                release_file = '{}/{}{}'.format(notifications_dir, '' if not tray else 'tray_', latest['tag_name'])
                if os.path.exists(release_file):
                    logger.info("Release {} already notified".format(latest['tag_name']))
                elif normalize_version(latest['tag_name']) > normalize_version(__version__):
                    try:
                        Path(notifications_dir).mkdir(parents=True, exist_ok=True)
                        with open(release_file, 'w+') as f:
                            f.write('')
                    except Exception:
                        logger.error("An error occurred while trying to create the update notification file: {}".format(release_file))

                    if tray:
                        return i18n['tray.warning.update_available'].format(__app_name__, latest['tag_name'])
                    else:
                        return i18n['warning.update_available'].format(bold(__app_name__), bold(latest['tag_name']), link(latest.get('html_url', '?')))
                else:
                    logger.info("No updates available")
            else:
                logger.warning("No official release found")
        else:
            logger.warning("No releases returned from the GitHub API")
    except Exception:
        logger.error("An error occurred while trying to retrieve the current releases")



import os
import traceback
from datetime import datetime, timedelta
from logging import Logger
from pathlib import Path
from threading import Thread
from typing import Optional, Dict

from bauh.api.abstract.handler import TaskManager
from bauh.api.abstract.model import SuggestionPriority
from bauh.api.http import HttpClient
from bauh.commons.boot import CreateConfigFile
from bauh.gems.arch import ARCH_CACHE_DIR, get_icon_path
from bauh.view.util.translation import I18n
from bauh.commons.suggestions import parse


class RepositorySuggestionsDownloader(Thread):

    _file_suggestions: Optional[str] = None
    _file_suggestions_ts: Optional[str] = None

    @classmethod
    def file_suggestions(cls) -> str:
        if cls._file_suggestions is None:
            cls._file_suggestions = f'{ARCH_CACHE_DIR}/suggestions.txt'

        return cls._file_suggestions

    @classmethod
    def file_suggestions_timestamp(cls) -> str:
        if cls._file_suggestions_ts is None:
            cls._file_suggestions_ts = f'{cls.file_suggestions()}.ts'

        return cls._file_suggestions_ts

    def __init__(self, logger: Logger, http_client: HttpClient, i18n: I18n,
                 create_config: Optional[CreateConfigFile] = None, file_url: Optional[str] = None):
        super(RepositorySuggestionsDownloader, self).__init__()
        self._log = logger
        self.i18n = i18n
        self.http_client = http_client
        self._taskman: Optional[TaskManager] = None
        self.create_config = create_config
        self._file_url = file_url if file_url else 'https://raw.githubusercontent.com/vinifmor/bauh-files' \
                                                   '/master/arch/suggestions.txt'
        self.task_id = 'arch.suggs'

    def register_task(self, taskman: Optional[TaskManager]):
        self._taskman = taskman
        if taskman:
            self._taskman.register_task(id_=self.task_id, label=self.i18n['task.download_suggestions'],
                                        icon_path=get_icon_path())

    @property
    def taskman(self) -> TaskManager:
        if self._taskman is None:
            self._taskman = TaskManager()

        return self._taskman

    def should_download(self, arch_config: dict, only_positive_exp: bool = False) -> bool:
        if not self._file_url:
            self._log.error("No Arch suggestions file URL defined")
            return False

        if self._file_url.startswith('/'):
            return False

        try:
            exp_hours = int(arch_config['suggestions_exp'])
        except ValueError:
            self._log.error(f"The Arch configuration property 'suggestions_exp' has a non int value set: "
                            f"{arch_config['suggestions']['expiration']}")
            return not only_positive_exp

        if exp_hours <= 0:
            self._log.info("Suggestions cache is disabled")
            return not only_positive_exp

        if not os.path.exists(self.file_suggestions()):
            self._log.info(f"'{self.file_suggestions()}' not found. It must be downloaded")
            return True

        if not os.path.exists(self.file_suggestions_timestamp()):
            self._log.info(f"'{self.file_suggestions()}' not found. The suggestions file must be downloaded.")
            return True

        with open(self.file_suggestions_timestamp()) as f:
            timestamp_str = f.read()

        try:
            suggestions_timestamp = datetime.fromtimestamp(float(timestamp_str))
        except Exception:
            self._log.error(f'Could not parse the Arch cached suggestions timestamp: {timestamp_str}')
            traceback.print_exc()
            return True

        update = suggestions_timestamp + timedelta(hours=exp_hours) <= datetime.utcnow()

        if update:
            self._log.info("The cached suggestions file is no longer valid")
        else:
            self._log.info("The cached suggestions file is up-to-date")

        return update

    def _save(self, text: str, timestamp: float):
        self._log.info(f"Caching suggestions to '{self.file_suggestions()}'")

        cache_dir = os.path.dirname(self.file_suggestions())

        try:
            Path(cache_dir).mkdir(parents=True, exist_ok=True)
            cache_dir_ok = True
        except OSError:
            self._log.error(f"Could not create cache directory '{cache_dir}'")
            traceback.print_exc()
            cache_dir_ok = False

        if cache_dir_ok:
            try:
                with open(self.file_suggestions(), 'w+') as f:
                    f.write(text)
            except Exception:
                self._log.error(f"An exception happened while writing the file '{self.file_suggestions()}'")
                traceback.print_exc()

            try:
                with open(self.file_suggestions_timestamp(), 'w+') as f:
                    f.write(str(timestamp))
            except Exception:
                self._log.error(f"An exception happened while writing the file '{self.file_suggestions_timestamp()}'")
                traceback.print_exc()

    def read_cached(self, custom_file: Optional[str] = None) -> Optional[Dict[str, SuggestionPriority]]:
        if custom_file:
            file_path, log_ref = custom_file, 'local'
        else:
            file_path, log_ref = self.file_suggestions(), 'cached'

        self._log.info(f"Reading {log_ref} Arch suggestions file '{file_path}'")

        try:
            with open(file_path) as f:
                sugs_str = f.read()
        except FileNotFoundError:
            self._log.warning(f"{log_ref.capitalize()} suggestions file does not exist ({file_path})")
            return

        if not sugs_str:
            self._log.warning(f"{log_ref.capitalize()} suggestions file '{file_path}' is empty")
            return

        return parse(sugs_str, self._log, 'Arch')

    def download(self) -> Optional[Dict[str, SuggestionPriority]]:
        self.taskman.update_progress(self.task_id, progress=1, substatus=None)

        self._log.info(f"Downloading suggestions from {self._file_url}")
        res = self.http_client.get(self._file_url)

        suggestions = None
        if res.status_code == 200 and res.text:
            self.taskman.update_progress(self.task_id, progress=50, substatus=None)
            suggestions = parse(res.text, self._log, 'Arch')

            if suggestions:
                self._save(text=res.text, timestamp=datetime.utcnow().timestamp())
            else:
                self._log.warning(f"Could not parse any Arch suggestion from {self._file_suggestions_ts}")
        else:
            self._log.warning(f"Could not retrieve Arch suggestions. "
                              f"Response (status={res.status_code}, text={res.text})")

        self.taskman.update_progress(self.task_id, progress=100, substatus=None)
        self.taskman.finish_task(self.task_id)
        return suggestions

    def read(self, arch_config: dict) -> Optional[Dict[str, int]]:
        if self._file_url:
            if self.is_custom_local_file_mapped():
                return self.read_cached(custom_file=self._file_url)

            if self.should_download(arch_config=arch_config):
                return self.download()

            return self.read_cached()

    def is_custom_local_file_mapped(self) -> bool:
        return self._file_url and self._file_url.startswith('/')

    def run(self):
        if self.create_config:
            if self.create_config.is_alive():
                self.taskman.update_progress(self.task_id, 0,
                                             self.i18n['task.waiting_task'].format(self.create_config.task_name))
                self.create_config.join()

            if not self.should_download(arch_config=self.create_config.config, only_positive_exp=False):
                self.taskman.update_progress(self.task_id, 100, self.i18n['task.canceled'])
                self.taskman.finish_task(self.task_id)
                return

            self.download()
        else:
            self._log.error(f"No {CreateConfigFile.__class__.__name__} instance set. Aborting..")
            self.taskman.update_progress(self.task_id, 100, self.i18n['error'])
            self.taskman.finish_task(self.task_id)



from bauh.commons.config import YAMLConfigManager
from bauh.gems.appimage import CONFIG_FILE


class AppImageConfigManager(YAMLConfigManager):

    def __init__(self):
        super(AppImageConfigManager, self).__init__(config_file_path=CONFIG_FILE)

    def get_default_config(self) -> dict:
        return {
            'database': {
                'expiration': 60
            },
            'suggestions': {
                'expiration': 24
            }
        }



import re
import time
from contextlib import contextmanager
from enum import Enum
from logging import Logger
from math import ceil
from queue import Queue
from threading import Thread
from typing import Iterable, Optional, Pattern, Dict, Set, Tuple, Generator, Collection

from bauh.api.abstract.handler import ProcessWatcher
from bauh.commons import system
from bauh.commons.html import bold
from bauh.commons.system import SimpleProcess
from bauh.commons.util import size_to_byte
from bauh.gems.debian.common import strip_maintainer_email, strip_section
from bauh.gems.debian.model import DebianPackage, DebianTransaction
from bauh.view.util.translation import I18n


def map_package_name(string: str):
    name_split = string.split(':')

    if len(name_split) > 2:
        return ':'.join(name_split[0:-1])

    return name_split[0]


class AptitudeAction(Enum):
    INSTALL = 0
    UPGRADE = 1
    REMOVE = 3


class Aptitude:

    def __init__(self, logger: Logger):
        self._log = logger
        self._re_show_attr: Optional[Pattern] = None
        self._env: Optional[Dict[str, str]] = None
        self._list_attrs: Optional[Set[str]] = None
        self._re_to_install: Optional[Pattern] = None
        self._re_transaction_pkg: Optional[Pattern] = None
        self._size_attrs: Optional[Tuple[str]] = None
        self._default_lang = ''
        self._ignored_fields: Optional[Set[str]] = None
        self._re_none: Optional[Pattern] = None
        self._vars_fixes: Optional[Dict[str, str]] = None
        self._preserve_env = {'DEBIAN_FRONTEND'}

    def show(self, pkgs: Iterable[str], attrs: Optional[Collection[str]] = None, verbose: bool = False) \
            -> Optional[Dict[str, Dict[str, object]]]:

        if pkgs:
            force_verbose = verbose if verbose else ('compressed size' in attrs if attrs else False)
            code, output = system.execute(f"aptitude show -q {' '.join(pkgs)}{' -v' if force_verbose else ''}",
                                          shell=True, custom_env=self.env)

            if code == 0 and output:
                info, pkg = dict(), None
                for field, val in self.re_show_attr.findall('\n' + output):
                    final_field, final_val = field.strip().lower(), val.strip()

                    if final_field == 'package':
                        pkg = {}
                        info[final_val] = pkg
                    elif final_field not in self.ignored_fields and (not attrs or final_field in attrs):
                        if final_val:
                            if final_field in self.list_attrs:
                                final_val = tuple((v.strip() for v in final_val.split(',') if v))
                            elif final_field in self.size_attrs:
                                size_split = final_val.split(' ')

                                if len(size_split) >= 1:
                                    unit = size_split[1].upper() if len(size_split) >= 2 else 'B'
                                    final_val = size_to_byte(size_split[0], unit, self._log)
                                else:
                                    self._log.warning(f"Unhandled value ({val}) for attribute '{field}'")
                                    final_val = None

                        pkg[final_field] = final_val

                return info

    def simulate_removal(self, packages: Iterable[str], purge: bool = False) -> Optional[DebianTransaction]:
        code, output = system.execute(self.gen_remove_cmd(packages, purge, simulate=True), shell=True,
                                      custom_env=self.env)

        if code == 0 and output:
            return self.map_transaction_output(output)

    def map_transaction_output(self, output: str) -> DebianTransaction:
        to_install, to_upgrade, to_remove = None, None, None
        current_collection = None

        for line in output.split('\n'):
            if line.startswith('The following NEW packages will be installed:'):
                to_install = set()
                current_collection = to_install
            elif line.startswith('The following packages will be upgraded:'):
                to_upgrade = set()
                current_collection = to_upgrade
            elif line.startswith('The following packages will be REMOVED:'):
                to_remove = set()
                current_collection = to_remove
            elif line.startswith('Would download/install/remove packages'):
                break
            elif current_collection is not None and line.startswith(' '):
                for n, _, v, __, lv, ___, size in self.re_transaction_pkg.findall(line):
                    pkg = DebianPackage(name=n, version=v, latest_version=lv if lv else v, transaction_size=0)

                    if size:
                        size_split = size.strip().split(' ')
                        unit = size_split[1][0].upper() if len(size_split) >= 2 else 'B'
                        pkg.transaction_size = size_to_byte(size_split[0], unit, self._log)

                    current_collection.add(pkg)

        return DebianTransaction(to_install=tuple(to_install) if to_install else tuple(),
                                 to_remove=tuple(to_remove) if to_remove else tuple(),
                                 to_upgrade=tuple(to_upgrade) if to_upgrade else tuple())

    def simulate_upgrade(self, packages: Iterable[str]) -> DebianTransaction:
        code, output = system.execute(self.gen_transaction_cmd('upgrade', packages, simulate=True),
                                      shell=True, custom_env=self.env)

        if code == 0 and output:
            return self.map_transaction_output(output)

    def upgrade(self, packages: Iterable[str], root_password: Optional[str]) -> SimpleProcess:
        cmd = self.gen_transaction_cmd('upgrade', packages).split(' ')
        return SimpleProcess(cmd=cmd, shell=True, root_password=root_password, extra_env=self.vars_fixes,
                             preserve_env=self._preserve_env)

    def update(self, root_password: Optional[str]) -> SimpleProcess:
        return SimpleProcess(('aptitude', 'update'), root_password=root_password, shell=True)

    def simulate_installation(self, packages: Iterable[str]) -> Optional[DebianTransaction]:
        code, output = system.execute(self.gen_transaction_cmd('install', packages, simulate=True),
                                      shell=True, custom_env=self.env)

        if code == 0 and output:
            return self.map_transaction_output(output)

    def install(self, packages: Iterable[str], root_password: Optional[str]) -> SimpleProcess:
        cmd = self.gen_transaction_cmd('install', packages).split(' ')
        return SimpleProcess(cmd=cmd, root_password=root_password, extra_env=self.vars_fixes,
                             preserve_env=self._preserve_env)

    def read_installed(self) -> Generator[DebianPackage, None, None]:
        yield from self.search(query='~i')

    def read_updates(self) -> Generator[Tuple[str, str], None, None]:
        _, output = system.execute(f"aptitude search ~U -q -F '%p^%V' --disable-columns --no-gui",
                                   shell=True,
                                   custom_env=self.env)

        if output:
            for line in output.split('\n'):
                line_split = line.strip().split('^')

                if len(line_split) == 2 and line_split[1] != '<none>':
                    yield line_split[0], line_split[1]

    def search(self, query: str, fill_size: bool = False) -> Generator[DebianPackage, None, None]:
        attrs = f"%p^%v^%V^%m^%s^{'%I^' if fill_size else ''}%d"
        _, output = system.execute(f"aptitude search {query} -q -F '{attrs}' --disable-columns", shell=True)

        if output:
            no_attrs = 7 if fill_size else 6

            for line in output.split('\n'):
                line_split = line.strip().split('^', maxsplit=no_attrs - 1)

                if len(line_split) == no_attrs:
                    latest_version = line_split[2] if not self.re_none.match(line_split[2]) else None

                    size = None

                    if fill_size:
                        size_split = line_split[no_attrs - 2].split(' ')
                        unit = size_split[1][0].upper() if len(size_split) >= 2 else 'B'
                        size = size_to_byte(size_split[0], unit, self._log)

                    if latest_version is not None:
                        installed_version = line_split[1] if not self.re_none.match(line_split[1]) else None
                        section = strip_section(line_split[4])

                        yield DebianPackage(name=line_split[0],
                                            version=installed_version if installed_version else latest_version,
                                            latest_version=latest_version,
                                            installed=bool(installed_version),
                                            update=installed_version is not None and installed_version != latest_version,
                                            maintainer=strip_maintainer_email(line_split[3]),
                                            categories=(section,) if section else None,
                                            uncompressed_size=size,
                                            description=line_split[no_attrs - 1])

    def search_by_name(self, names: Iterable[str], fill_size: bool = False) -> Generator[DebianPackage, None, None]:
        query = f"'({'|'.join(f'?exact-name({n})' for n in names)})'"
        yield from self.search(query=query, fill_size=fill_size)

    def remove(self, packages: Iterable[str], root_password: Optional[str], purge: bool = False) -> SimpleProcess:
        return SimpleProcess(cmd=self.gen_remove_cmd(packages, purge).split(' '), shell=True,
                             root_password=root_password, extra_env=self.vars_fixes, preserve_env=self._preserve_env)

    def read_installed_names(self) -> Generator[str, None, None]:
        code, output = system.execute("aptitude search ~i -q -F '%p' --disable-columns",
                                      shell=True,
                                      custom_env=self.env)

        if output:
            for line in output.split('\n'):
                if line:
                    yield line

    @property
    def re_show_attr(self) -> Pattern:
        if self._re_show_attr is None:
            self._re_show_attr = re.compile(r'(\n\w+[a-zA-Z0-9\-\s]*):\s+(.+)')

        return self._re_show_attr

    @property
    def env(self) -> Dict[str, str]:
        if self._env is None:
            self._env = system.gen_env(global_interpreter=system.USE_GLOBAL_INTERPRETER)
            self._env['LC_NUMERIC'] = ''

        return self._env

    @property
    def list_attrs(self) -> Set[str]:
        if self._list_attrs is None:
            self._list_attrs = {'depends', 'provides', 'replaces', 'recommends', 'suggests', 'conflicts',
                                'state', 'predepends', 'breaks'}

        return self._list_attrs

    @property
    def re_transaction_pkg(self) -> Pattern:
        if self._re_transaction_pkg is None:
            self._re_transaction_pkg = re.compile(r'([a-zA-Z0-9\-_@~.+:]+)({\w+})?\s*\[([a-zA-Z0-9\-_@~.+:]+)'
                                                  r'(\s+->\s+([a-zA-Z0-9\-_@~.+:]+))?](\s*<([\-+]?[0-9.,]+\s+\w+)>)?')

        return self._re_transaction_pkg

    @property
    def size_attrs(self) -> Tuple[str]:
        if self._size_attrs is None:
            self._size_attrs = {'compressed size', 'uncompressed size'}

        return self._size_attrs

    @property
    def ignored_fields(self) -> Set[str]:
        if self._ignored_fields is None:
            self._ignored_fields = {'sha1', 'sha256', 'sha512', 'checksum-filesize'}

        return self._ignored_fields

    @classmethod
    def gen_remove_cmd(cls, packages: Iterable[str], purge: bool, simulate: bool = False) -> str:
        return cls.gen_transaction_cmd(type_='purge' if purge else 'remove', packages=packages,
                                       simulate=simulate)

    @staticmethod
    def gen_transaction_cmd(type_: str, packages: Iterable[str], simulate: bool = False,
                            delete_unused: bool = False) -> str:
        return f"aptitude {type_} -q -y --no-gui --full-resolver {' '.join(packages)}" \
               f" -o Aptitude::ProblemResolver::RemoveScore=9999999" \
               f" -o Aptitude::ProblemResolver::EssentialRemoveScore=9999999" \
               f" -o Aptitude::Delete-Unused={str(delete_unused).lower()}" \
               f"{' -V -s -Z' if simulate else ''}"

    @property
    def re_none(self) -> Pattern:
        if self._re_none is None:
            self._re_none = re.compile(r'^<\w+>$')

        return self._re_none

    @property
    def vars_fixes(self) -> Dict[str, str]:
        if self._vars_fixes is None:
            self._vars_fixes = {'LC_NUMERIC': '', 'DEBIAN_FRONTEND': 'noninteractive'}

        return self._vars_fixes


class AptitudeOutputHandler(Thread):

    def __init__(self, i18n: I18n, targets: Iterable[str], re_download: Pattern, watcher: ProcessWatcher,
                 action: AptitudeAction):
        super(AptitudeOutputHandler, self).__init__()
        self._i18n = i18n
        self._re_download = re_download
        self._watcher = watcher
        self._targets = set(targets) if targets is not None else None
        self._unpacking = 0
        self._removing = 0
        self._downloading = 0
        self._to_process = Queue()
        self._work = True
        self._action = action

    def stop_working(self):
        self._work = False

    def handle(self, string: str):
        self._to_process.put(string)

    @property
    def total_targets(self) -> int:
        return len(self._targets) if self._targets else 0

    @property
    def processed(self) -> int:
        return self._removing if self._action == AptitudeAction.REMOVE else self._unpacking

    def _get_progress(self, current: int) -> str:
        if self.total_targets > 0:
            if self._action == AptitudeAction.REMOVE:
                total = self._removing
            else:
                if self.total_targets == self._unpacking:
                    total = self.total_targets
                else:
                    total = ceil((self._unpacking + self._downloading) / 2)

            return f'({total / self.total_targets * 100:.2f}%) [{current}/{self.total_targets}] '

        return ''

    def run(self):
        while self._work:
            time.sleep(0.001)

            if self._to_process.empty():
                continue

            string = self._to_process.get()

            if self.total_targets > 0 and self.total_targets == self._unpacking:
                self._watcher.change_substatus(self._i18n['debian.output.finishing'])
                continue

            if string:
                if self._action != AptitudeAction.REMOVE and string.startswith('Unpacking '):
                    unpacking = string.split(' ')

                    if len(unpacking) >= 2 and unpacking[1]:
                        pkg = map_package_name(unpacking[1].strip())

                        if self._targets and pkg in self._targets:
                            self._unpacking += 1

                        msg = f"{self._get_progress(self._unpacking)}" \
                              f"{self._i18n['debian.output.unpacking'].format(pkg=bold(pkg))}"

                        self._watcher.change_substatus(msg)

                        continue

                if self._action == AptitudeAction.REMOVE and string.startswith('Removing '):
                    unpacking = string.split(' ')

                    if len(unpacking) >= 2 and unpacking[1]:
                        pkg = unpacking[1].strip()

                        if self._targets and pkg in self._targets:
                            self._removing += 1

                        msg = f"{self._get_progress(self._removing)}" \
                              f"{self._i18n['debian.output.removing'].format(pkg=bold(pkg))}"

                        self._watcher.change_substatus(msg)

                        continue

                download = self._re_download.findall(string)

                if download:
                    data = download[0].split(' ')

                    if len(data) >= 4:
                        pkg = data[3].strip()

                        if self._targets and pkg in self._targets:
                            self._downloading += 1

                        msg = f"{self._get_progress(self._downloading)}" \
                              f"{self._i18n['debian.output.downloading'].format(pkg=bold(pkg))}"

                        self._watcher.change_substatus(msg)

                        continue

                _processed = self.processed
                if self._targets and _processed > 0:
                    self._watcher.change_substatus(self._get_progress(_processed).strip())
                    continue

            self._watcher.change_substatus(' ')


class AptitudeOutputHandlerFactory:

    def __init__(self, i18n: I18n):
        self._i18n = i18n
        self._re_download: Optional[Pattern] = None

    @property
    def re_download(self) -> Pattern:
        if self._re_download is None:
            self._re_download = re.compile(r'Get:\s+\d+\s+https?://(.+)')

        return self._re_download

    @contextmanager
    def start(self, watcher: ProcessWatcher, targets: Iterable[str], action: AptitudeAction):
        handler = AptitudeOutputHandler(i18n=self._i18n, re_download=self.re_download,
                                        watcher=watcher, targets=targets, action=action)
        handler.start()

        yield handler.handle

        handler.stop_working()
        handler.join()



import os
import shutil
from pathlib import Path
from typing import Set

from bauh.commons import system
from bauh.gems.arch import IGNORED_REBUILD_CHECK_FILE


def is_installed() -> bool:
    return bool(shutil.which('checkrebuild'))


def list_required_rebuild() -> Set[str]:
    code, output = system.execute(cmd='checkrebuild', shell=True, stdin=False)

    required = set()
    if code == 0 and output:
        for line in output.split('\n'):
            line_strip = line.strip()

            if line_strip:
                line_split = line_strip.split('\t')

                if len(line_split) > 1:
                    required.add(line_split[1])

    return required


def list_ignored() -> Set[str]:
    if os.path.isfile(IGNORED_REBUILD_CHECK_FILE):
        with open(IGNORED_REBUILD_CHECK_FILE) as f:
            ignored_str = f.read()

        return {p.strip() for p in ignored_str.split('\n') if p}
    else:
        return set()


def add_as_ignored(pkgname: str):
    ignored = list_ignored()
    ignored.add(pkgname)

    Path(os.path.dirname(IGNORED_REBUILD_CHECK_FILE)).mkdir(parents=True, exist_ok=True)

    with open(IGNORED_REBUILD_CHECK_FILE, 'w+') as f:
        for p in ignored:
            f.write('{}\n'.format(p))


def remove_from_ignored(pkgname: str):
    ignored = list_ignored()

    if ignored is None or pkgname not in ignored:
        return

    ignored.remove(pkgname)

    Path(os.path.dirname(IGNORED_REBUILD_CHECK_FILE)).mkdir(parents=True, exist_ok=True)

    if ignored:
        with open(IGNORED_REBUILD_CHECK_FILE, 'w+') as f:
            for p in ignored:
                f.write('{}\n'.format(p))
    else:
        os.remove(IGNORED_REBUILD_CHECK_FILE)



import logging
import sys
from typing import Optional, Dict

from bauh.api.abstract.cache import MemoryCacheFactory
from bauh.api.abstract.disk import DiskCacheLoaderFactory
from bauh.api.abstract.download import FileDownloader
from bauh.api.http import HttpClient
from bauh.commons.internet import InternetChecker
from bauh.view.util.translation import I18n


class ApplicationContext:

    def __init__(self, download_icons: bool, http_client: HttpClient, app_root_dir: str, i18n: I18n,
                 cache_factory: MemoryCacheFactory, disk_loader_factory: DiskCacheLoaderFactory,
                 logger: logging.Logger, file_downloader: FileDownloader, distro: str, app_name: str,
                 app_version: str, internet_checker: InternetChecker, root_user: bool, screen_width: int = -1,
                 screen_height: int = -1, suggestions_mapping: Optional[Dict[str, str]] = None):
        """
        :param download_icons: if packages icons should be downloaded
        :param http_client: a shared instance of http client
        :param app_root_dir: GUI root dir
        :param i18n: the translation keys
        :param cache_factory:
        :param disk_loader_factory:
        :param logger: a logger instance
        :param file_downloader
        :param distro
        :param app_name
        :param app_version
        :param internet_checker
        :param screen_width
        :param screen_height
        :param suggestions_mapping
        """
        self.download_icons = download_icons
        self.http_client = http_client
        self.app_root_dir = app_root_dir
        self.i18n = i18n
        self.cache_factory = cache_factory
        self.disk_loader_factory = disk_loader_factory
        self.logger = logger
        self.file_downloader = file_downloader
        self.arch_x86_64 = sys.maxsize > 2**32
        self.distro = distro
        self.default_categories = ('AudioVideo', 'Audio', 'Video', 'Development', 'Education', 'Game',
                                   'Graphics', 'Network', 'Office', 'Science', 'Settings', 'System', 'Utility')
        self.app_name = app_name
        self.app_version = app_version
        self.root_user = root_user
        self.root_password = None
        self.internet_checker = internet_checker
        self.screen_width = screen_width
        self.screen_height = screen_height
        self._suggestions_mapping = suggestions_mapping

    def is_system_x86_64(self):
        return self.arch_x86_64

    def get_view_path(self):
        return self.app_root_dir + '/view'

    def is_internet_available(self) -> bool:
        return self.internet_checker.is_available()

    def get_suggestion_url(self, module: str, default: Optional[str] = None) -> Optional[str]:
        if self._suggestions_mapping:
            module_split = module.split(f'{self.app_name}.gems.')

            if len(module_split) > 1:
                gem_name = module_split[1].split('.')[0]
                return self._suggestions_mapping.get(gem_name, default)

        return default



from logging import Logger
from typing import Dict, Optional, Tuple

from bauh.api.abstract.model import SuggestionPriority


def parse(suggestions_str: str, logger: Optional[Logger] = None, type_: Optional[str] = None,
          splitter: str = '=') \
        -> Dict[str, SuggestionPriority]:
    output = dict()

    for line in suggestions_str.split('\n'):
        clean_line = line.strip()

        if clean_line:
            line_split = clean_line.split(splitter, 1)

            if len(line_split) == 2:
                prio_str, name = line_split[0].strip(), line_split[1].strip()

                if prio_str and name:
                    try:
                        prio = int(line_split[0])
                    except ValueError:
                        if logger:
                            logger.warning(f"Could not parse {type_ + ' ' if type_ else ''}suggestion: {line}")
                        continue

                    output[line_split[1]] = SuggestionPriority(prio)

    return output


def sort_by_priority(names_prios: Dict[str, SuggestionPriority]) -> Tuple[str, ...]:
    return tuple(pair[1] for pair in sorted(((names_prios[n], n) for n in names_prios), reverse=True))



import os
import re
import shutil
import time
import traceback
from subprocess import Popen, STDOUT
from threading import Thread
from typing import List, Set, Type, Tuple, Dict, Optional, Generator, Callable, Pattern

from bauh.api.abstract.controller import SoftwareManager, SearchResult, ApplicationContext, UpgradeRequirements, \
    UpgradeRequirement, TransactionResult, SoftwareAction, SettingsView, SettingsController
from bauh.api.abstract.disk import DiskCacheLoader
from bauh.api.abstract.handler import ProcessWatcher, TaskManager
from bauh.api.abstract.model import SoftwarePackage, PackageUpdate, PackageHistory, PackageSuggestion, \
    CustomSoftwareAction
from bauh.api.abstract.view import ViewComponent, TabGroupComponent, MessageType, PanelComponent
from bauh.api.exception import NoInternetException
from bauh.commons.boot import CreateConfigFile
from bauh.commons.html import bold
from bauh.commons.util import sanitize_command_input
from bauh.view.core.config import CoreConfigManager
from bauh.view.core.settings import GenericSettingsManager
from bauh.view.core.update import check_for_update
from bauh.view.util import resource
from bauh.view.util.resource import get_path
from bauh.view.util.util import clean_app_files, restart_app

RE_IS_URL = re.compile(r'^https?://.+')


class GenericUpgradeRequirements(UpgradeRequirements):

    def __init__(self, to_install: List[UpgradeRequirement], to_remove: List[UpgradeRequirement],
                 to_upgrade: List[UpgradeRequirement], cannot_upgrade: List[SoftwarePackage],
                 sub_requirements: Dict[SoftwareManager, UpgradeRequirements]):
        super(GenericUpgradeRequirements, self).__init__(to_install=to_install, to_upgrade=to_upgrade,
                                                         to_remove=to_remove, cannot_upgrade=cannot_upgrade)
        self.sub_requirements = sub_requirements


class GenericSoftwareManager(SoftwareManager, SettingsController):

    def __init__(self, managers: List[SoftwareManager], context: ApplicationContext, config: dict,
                 force_suggestions: bool = False):

        super(GenericSoftwareManager, self).__init__(context=context)
        self.managers = managers
        self.map = {t: m for m in self.managers for t in m.get_managed_types()}
        self._available_cache = {} if config['system']['single_dependency_checking'] else None
        self.thread_prepare = None
        self.i18n = context.i18n
        self.disk_loader_factory = context.disk_loader_factory
        self.logger = context.logger
        self._already_prepared = []
        self.working_managers = []
        self.config = config
        self.settings_manager: Optional[GenericSettingsManager] = None
        self.http_client = context.http_client
        self.configman = CoreConfigManager()
        self._action_reset: Optional[CustomSoftwareAction] = None
        self._dynamic_extra_actions: Optional[Dict[CustomSoftwareAction, Callable[[dict], bool]]] = None
        self.force_suggestions = force_suggestions

    @property
    def dynamic_extra_actions(self) -> Dict[CustomSoftwareAction, Callable[[dict], bool]]:
        if self._dynamic_extra_actions is None:
            self._dynamic_extra_actions = {
                CustomSoftwareAction(i18n_label_key='action.backups',
                                     i18n_status_key='action.backups.status',
                                     i18n_description_key='action.backups.desc',
                                     manager_method='launch_timeshift',
                                     manager=self,
                                     icon_path='timeshift',
                                     requires_root=False,
                                     refresh=False): self.is_backups_action_available
            }

        return self._dynamic_extra_actions

    @property
    def action_reset(self) -> CustomSoftwareAction:
        if self._action_reset is None:
            self._action_reset = CustomSoftwareAction(i18n_label_key='action.reset',
                                                      i18n_status_key='action.reset.status',
                                                      i18n_description_key='action.reset.desc',
                                                      manager_method='reset',
                                                      icon_path=resource.get_path('img/logo.svg'),
                                                      requires_root=False,
                                                      manager=self,
                                                      refresh=False)

        return self._action_reset

    def _is_timeshift_launcher_available(self) -> bool:
        return bool(shutil.which('timeshift-launcher'))

    def is_backups_action_available(self, app_config: dict) -> bool:
        return bool(app_config['backup']['enabled']) and self._is_timeshift_launcher_available()

    def reset_cache(self):
        if self._available_cache is not None:
            self._available_cache = {}
            self.working_managers.clear()

    def launch_timeshift(self, root_password: Optional[str], watcher: ProcessWatcher):
        if self._is_timeshift_launcher_available():
            try:
                Popen(['timeshift-launcher'], stderr=STDOUT)
                return True
            except Exception:
                traceback.print_exc()
                watcher.show_message(title=self.i18n["error"].capitalize(),
                                     body=self.i18n['action.backups.tool_error'].format(bold('Timeshift')),
                                     type_=MessageType.ERROR)
                return False
        else:
            watcher.show_message(title=self.i18n["error"].capitalize(),
                                 body=self.i18n['action.backups.tool_error'].format(bold('Timeshift')),
                                 type_=MessageType.ERROR)
            return False

    def _can_work(self, man: SoftwareManager):
        if self._available_cache is not None:
            available = False
            for t in man.get_managed_types():
                available = self._available_cache.get(t)

                if available is None:
                    available = man.is_enabled() and man.can_work()[0]
                    self._available_cache[t] = available

                if available:
                    available = True
        else:
            available = man.is_enabled() and man.can_work()[0]

        if available:
            if man not in self.working_managers:
                self.working_managers.append(man)
        else:
            if man in self.working_managers:
                self.working_managers.remove(man)

        return available

    def _search(self, word: str, is_url: bool, man: SoftwareManager, disk_loader, res: SearchResult):
        if self._can_work(man):
            mti = time.time()
            apps_found = man.search(words=word, disk_loader=disk_loader, is_url=is_url, limit=-1)
            mtf = time.time()
            self.logger.info(f'{man.__class__.__name__} took {mtf - mti:.8f} seconds')

            res.installed.extend(apps_found.installed)
            res.new.extend(apps_found.new)

    def search(self, words: str, disk_loader: DiskCacheLoader = None, limit: int = -1, is_url: bool = False) -> SearchResult:
        ti = time.time()
        self._wait_to_be_ready()

        res = SearchResult.empty()

        if self.context.is_internet_available():
            norm_query = sanitize_command_input(words).lower()
            self.logger.info(f"Search query: {norm_query}")

            if norm_query:
                is_url = bool(RE_IS_URL.match(norm_query))
                disk_loader = self.disk_loader_factory.new()
                disk_loader.start()

                threads = []

                for man in self.managers:
                    t = Thread(target=self._search, args=(norm_query, is_url, man, disk_loader, res))
                    t.start()
                    threads.append(t)

                for t in threads:
                    t.join()

                if disk_loader:
                    disk_loader.stop_working()
                    disk_loader.join()

            # res.installed = self._sort(res.installed, norm_word)
            # res.new = self._sort(res.new, norm_word)
        else:
            raise NoInternetException()

        res.update_total()
        tf = time.time()
        self.logger.info(f'Took {tf - ti:.8f} seconds')
        return res

    def _wait_to_be_ready(self):
        if self.thread_prepare:
            self.thread_prepare.join()
            self.thread_prepare = None

    def set_enabled(self, enabled: bool):
        pass

    def can_work(self) -> Tuple[bool, Optional[str]]:
        return True, None

    def _get_package_lower_name(self, pkg: SoftwarePackage):
        return pkg.name.lower()

    def read_installed(self, disk_loader: DiskCacheLoader = None, limit: int = -1, only_apps: bool = False, pkg_types: Set[Type[SoftwarePackage]] = None, internet_available: bool = None) -> SearchResult:
        ti = time.time()
        self._wait_to_be_ready()

        res = SearchResult([], None, 0)

        disk_loader = None

        net_available = self.context.is_internet_available()
        if not pkg_types:  # any type
            for man in self.managers:
                if self._can_work(man):
                    if not disk_loader:
                        disk_loader = self.disk_loader_factory.new()
                        disk_loader.start()

                    mti = time.time()
                    man_res = man.read_installed(disk_loader=disk_loader, pkg_types=None, internet_available=net_available)
                    mtf = time.time()
                    self.logger.info(f'{man.__class__.__name__} took {mtf - mti:.2f} seconds')

                    res.installed.extend(man_res.installed)
                    res.total += man_res.total
        else:
            man_already_used = []

            for t in pkg_types:
                man = self.map.get(t)
                if man and (man not in man_already_used) and self._can_work(man):

                    if not disk_loader:
                        disk_loader = self.disk_loader_factory.new()
                        disk_loader.start()

                    mti = time.time()
                    man_res = man.read_installed(disk_loader=disk_loader, pkg_types=None, internet_available=net_available)
                    mtf = time.time()
                    self.logger.info(f'{man.__class__.__name__} took {mtf - mti:.2f} seconds')

                    res.installed.extend(man_res.installed)
                    res.total += man_res.total

        if disk_loader:
            disk_loader.stop_working()
            disk_loader.join()

        if res.installed:
            for p in res.installed:
                if p.is_update_ignored():
                    if p.categories is None:
                        p.categories = ['updates_ignored']
                    elif 'updates_ignored' not in p.categories:
                        self._add_category(p, 'updates_ignored')

            res.installed.sort(key=self._get_package_lower_name)

        tf = time.time()
        self.logger.info(f'Took {tf - ti:.2f} seconds')
        return res

    def _add_category(self, pkg: SoftwarePackage, category: str):
        if isinstance(pkg.categories, tuple):
            pkg.categories = tuple((*pkg.categories, category))
        elif isinstance(pkg.categories, list):
            pkg.categories.append(category)
        elif isinstance(pkg.categories, set):
            pkg.categories.add(category)

    def downgrade(self, app: SoftwarePackage, root_password: Optional[str], handler: ProcessWatcher) -> bool:
        man = self._get_manager_for(app)

        if man and app.can_be_downgraded():
            mti = time.time()
            res = man.downgrade(app, root_password, handler)
            mtf = time.time()
            self.logger.info(f'Took {mtf - mti:.2f} seconds')
            return res
        else:
            raise Exception(f"Downgrading is not possible for {app.__class__.__name__}")

    def clean_cache_for(self, app: SoftwarePackage):
        man = self._get_manager_for(app)

        if man:
            return man.clean_cache_for(app)

    def upgrade(self, requirements: GenericUpgradeRequirements, root_password: Optional[str], handler: ProcessWatcher) -> bool:
        for man, man_reqs in requirements.sub_requirements.items():
            res = man.upgrade(man_reqs, root_password, handler)

            if not res:
                return False

        return True

    def _fill_post_transaction_status(self, pkg: SoftwarePackage, installed: bool):
        pkg.installed = installed
        pkg.update = False

        if pkg.latest_version:
            pkg.version = pkg.latest_version

    def _update_post_transaction_status(self, res: TransactionResult):
        if res.success:
            if res.installed:
                for p in res.installed:
                    self._fill_post_transaction_status(p, True)
            if res.removed:
                for p in res.removed:
                    self._fill_post_transaction_status(p, False)

    def uninstall(self, pkg: SoftwarePackage, root_password: Optional[str], handler: ProcessWatcher, disk_loader: DiskCacheLoader = None) -> TransactionResult:
        man = self._get_manager_for(pkg)

        if man:
            ti = time.time()
            disk_loader = self.disk_loader_factory.new()
            disk_loader.start()
            self.logger.info(f"Uninstalling {pkg.name}")
            try:
                res = man.uninstall(pkg, root_password, handler, disk_loader)
                disk_loader.stop_working()
                disk_loader.join()
                self._update_post_transaction_status(res)
                return res
            except Exception:
                traceback.print_exc()
                return TransactionResult(success=False, installed=[], removed=[])
            finally:
                tf = time.time()
                self.logger.info(f'Uninstallation of {pkg} took {(tf - ti) / 60:.2f} minutes')

    def install(self, app: SoftwarePackage, root_password: Optional[str], disk_loader: DiskCacheLoader, handler: ProcessWatcher) -> TransactionResult:
        man = self._get_manager_for(app)

        if man:
            ti = time.time()
            disk_loader = self.disk_loader_factory.new()
            disk_loader.start()
            try:
                self.logger.info(f'Installing {app}')
                res = man.install(app, root_password, disk_loader, handler)
                disk_loader.stop_working()
                disk_loader.join()
                self._update_post_transaction_status(res)
                return res
            except Exception:
                traceback.print_exc()
                return TransactionResult(success=False, installed=[], removed=[])
            finally:
                tf = time.time()
                self.logger.info(f'Installation of {app} took {(tf - ti) / 60:.2f} minutes')

    def get_info(self, app: SoftwarePackage):
        man = self._get_manager_for(app)

        if man:
            return man.get_info(app)

    def get_history(self, app: SoftwarePackage) -> PackageHistory:
        man = self._get_manager_for(app)

        if man:
            mti = time.time()
            history = man.get_history(app)
            mtf = time.time()
            self.logger.info(f'{man.__class__.__name__} took {mtf - mti:.2f} seconds')
            return history

    def get_managed_types(self) -> Set[Type[SoftwarePackage]]:
        available_types = set()

        for man in self.get_working_managers():
            available_types.update(man.get_managed_types())

        return available_types

    def is_enabled(self):
        return True

    def _get_manager_for(self, app: SoftwarePackage) -> SoftwareManager:
        man = self.map[app.__class__]
        return man if man and self._can_work(man) else None

    def cache_to_disk(self, pkg: SoftwarePackage, icon_bytes: bytes, only_icon: bool):
        if pkg.supports_disk_cache():
            man = self._get_manager_for(pkg)

            if man:
                return man.cache_to_disk(pkg, icon_bytes=icon_bytes, only_icon=only_icon)

    def requires_root(self, action: SoftwareAction, app: SoftwarePackage) -> bool:
        if app is None:
            if self.managers:
                for man in self.managers:
                    if self._can_work(man):
                        if man.requires_root(action, app):
                            return True
            return False
        else:
            man = self._get_manager_for(app)

            if man:
                return man.requires_root(action, app)

    def prepare(self, task_manager: TaskManager, root_password: Optional[str], internet_available: bool):
        ti = time.time()
        self.logger.info("Initializing")
        taskman = task_manager if task_manager else TaskManager()  # empty task manager to prevent null pointers

        create_config = CreateConfigFile(taskman=taskman, configman=self.configman, i18n=self.i18n,
                                         task_icon_path=get_path('img/logo.svg'), logger=self.logger)
        create_config.start()

        if self.managers:
            internet_on = self.context.is_internet_available()
            prepare_tasks = []
            for man in self.managers:
                if man not in self._already_prepared and self._can_work(man):
                    t = Thread(target=man.prepare, args=(taskman, root_password, internet_on), daemon=True)
                    t.start()
                    prepare_tasks.append(t)
                    self._already_prepared.append(man)

            for t in prepare_tasks:
                t.join()

        create_config.join()
        tf = time.time()
        self.logger.info(f'Finished ({tf - ti:.2f} seconds)')

    def cache_available_managers(self):
        if self.managers:
            for man in self.managers:
                self._can_work(man)

    def list_updates(self, internet_available: bool = None) -> List[PackageUpdate]:
        self._wait_to_be_ready()

        updates = []

        if self.managers:
            net_available = self.context.is_internet_available()

            for man in self.managers:
                if self._can_work(man):
                    man_updates = man.list_updates(internet_available=net_available)
                    if man_updates:
                        updates.extend(man_updates)

        return updates

    def list_warnings(self, internet_available: bool = None) -> List[str]:
        warnings = []

        int_available = self.context.is_internet_available()

        if int_available:
            updates_msg = check_for_update(self.logger, self.http_client, self.i18n)

            if updates_msg:
                warnings.append(updates_msg)

        if self.managers:
            for man in self.managers:
                if self._can_work(man):
                    man_warnings = man.list_warnings(internet_available=int_available)

                    if man_warnings:
                        warnings.extend(man_warnings)

        return warnings

    def _fill_suggestions(self, suggestions: list, man: SoftwareManager, limit: int, filter_installed: bool):
        if self._can_work(man):
            mti = time.time()
            man_sugs = man.list_suggestions(limit=limit, filter_installed=filter_installed)
            mtf = time.time()
            self.logger.info(f'{man.__class__.__name__} took {mtf - mti:.5f} seconds')

            if man_sugs:
                if 0 < limit < len(man_sugs):
                    man_sugs = man_sugs[0:limit]

                suggestions.extend(man_sugs)

    def list_suggestions(self, limit: int, filter_installed: bool) -> List[PackageSuggestion]:
        if self.force_suggestions or bool(self.config['suggestions']['enabled']):
            if self.managers and self.context.is_internet_available():
                suggestions, threads = [], []
                for man in self.managers:
                    t = Thread(target=self._fill_suggestions,
                               args=(suggestions, man, int(self.config['suggestions']['by_type']), filter_installed))
                    t.start()
                    threads.append(t)

                for t in threads:
                    t.join()

                if suggestions:
                    suggestions.sort(key=lambda s: s.priority.value, reverse=True)

                return suggestions
        return []

    def execute_custom_action(self, action: CustomSoftwareAction, pkg: SoftwarePackage, root_password: Optional[str], watcher: ProcessWatcher):
        if action.requires_internet and not self.context.is_internet_available():
            raise NoInternetException()

        man = action.manager if action.manager else self._get_manager_for(pkg)

        if man:
            return eval(f"man.{action.manager_method}({'pkg=pkg, ' if pkg else ''}root_password=root_password, watcher=watcher)")

    def is_default_enabled(self) -> bool:
        return True

    def launch(self, pkg: SoftwarePackage):
        self._wait_to_be_ready()

        man = self._get_manager_for(pkg)

        if man:
            self.logger.info(f'Launching {pkg}')
            man.launch(pkg)

    def get_screenshots(self, pkg: SoftwarePackage) -> Generator[str, None, None]:
        man = self._get_manager_for(pkg)

        if man:
            yield from man.get_screenshots(pkg)

    def get_working_managers(self):
        return [m for m in self.managers if self._can_work(m)]

    def get_settings(self) -> Optional[Generator[SettingsView, None, None]]:
        if self.settings_manager is None:
            self.settings_manager = GenericSettingsManager(managers=self.managers,
                                                           working_managers=self.working_managers,
                                                           configman=self.configman,
                                                           context=self.context)
        else:
            self.settings_manager.managers = self.managers
            self.settings_manager.working_managers = self.working_managers

        yield SettingsView(self, self.settings_manager.get_settings())

    def save_settings(self, component: TabGroupComponent) -> Tuple[bool, Optional[List[str]]]:
        return self.settings_manager.save_settings(component)

    def _map_pkgs_by_manager(self, pkgs: List[SoftwarePackage], pkg_filters: list = None) -> Dict[SoftwareManager, List[SoftwarePackage]]:
        by_manager = {}
        for pkg in pkgs:
            if pkg_filters and not all((1 for f in pkg_filters if f(pkg))):
                continue

            man = self._get_manager_for(pkg)

            if man:
                man_pkgs = by_manager.get(man)

                if man_pkgs is None:
                    man_pkgs = []
                    by_manager[man] = man_pkgs

                man_pkgs.append(pkg)

        return by_manager

    def get_upgrade_requirements(self, pkgs: List[SoftwarePackage], root_password: Optional[str], watcher: ProcessWatcher) -> UpgradeRequirements:
        by_manager = self._map_pkgs_by_manager(pkgs)
        res = GenericUpgradeRequirements([], [], [], [], {})

        if by_manager:
            for man, pkgs in by_manager.items():
                ti = time.time()
                man_reqs = man.get_upgrade_requirements(pkgs, root_password, watcher)
                tf = time.time()
                self.logger.info(f'{man.__class__.__name__} took {tf - ti:.2f} seconds')

                if not man_reqs:
                    return  # it means the process should be stopped

                if man_reqs:
                    res.sub_requirements[man] = man_reqs
                    if man_reqs.to_install:
                        res.to_install.extend(man_reqs.to_install)

                    if man_reqs.to_remove:
                        res.to_remove.extend(man_reqs.to_remove)

                    if man_reqs.to_upgrade:
                        res.to_upgrade.extend(man_reqs.to_upgrade)

                    if man_reqs.cannot_upgrade:
                        res.cannot_upgrade.extend(man_reqs.cannot_upgrade)

        return res

    def reset(self, root_password: Optional[str], watcher: ProcessWatcher) -> bool:
        body = f"<p>{self.i18n['action.reset.body_1'].format(bold(self.context.app_name))}</p>" \
               f"<p>{self.i18n['action.reset.body_2']}</p>"

        if watcher.request_confirmation(title=self.i18n['action.reset'],
                                        body=body,
                                        confirmation_label=self.i18n['proceed'].capitalize(),
                                        deny_label=self.i18n['cancel'].capitalize()):

            try:
                clean_app_files(managers=self.managers, logs=False)
                restart_app()
            except Exception:
                return False

        return True

    def gen_custom_actions(self) -> Generator[CustomSoftwareAction, None, None]:
        if self.managers:
            working_managers = []

            for man in self.managers:
                if self._can_work(man):
                    working_managers.append(man)

            if working_managers:
                working_managers.sort(key=lambda m: m.__class__.__name__)

                for man in working_managers:
                    for action in man.gen_custom_actions():
                        action.manager = man
                        yield action

        app_config = self.configman.get_config()

        for action, available in self.dynamic_extra_actions.items():
            if available(app_config):
                yield action

        yield self.action_reset

    def _fill_sizes(self, man: SoftwareManager, pkgs: List[SoftwarePackage]):
        ti = time.time()
        man.fill_sizes(pkgs)
        tf = time.time()
        self.logger.info(f'{man.__class__.__name__} took {tf - ti:.2f} seconds')

    def fill_sizes(self, pkgs: List[SoftwarePackage]):
        by_manager = self._map_pkgs_by_manager(pkgs, pkg_filters=[lambda p: p.size is None])

        if by_manager:
            threads = []
            for man, man_pkgs in by_manager.items():
                if man_pkgs:
                    t = Thread(target=self._fill_sizes, args=(man, man_pkgs), daemon=True)
                    t.start()
                    threads.append(t)

            for t in threads:
                t.join()

    def ignore_update(self, pkg: SoftwarePackage):
        manager = self._get_manager_for(pkg)

        if manager:
            manager.ignore_update(pkg)

            if pkg.is_update_ignored():
                if pkg.categories is None:
                    pkg.categories = ['updates_ignored']
                elif 'updates_ignored' not in pkg.categories:
                    self._add_category(pkg, 'updates_ignored')

    def revert_ignored_update(self, pkg: SoftwarePackage):
        manager = self._get_manager_for(pkg)

        if manager:
            manager.revert_ignored_update(pkg)

            if not pkg.is_update_ignored() and pkg.categories and 'updates_ignored' in pkg.categories:
                if isinstance(pkg.categories, tuple):
                    pkg.categories = tuple(c for c in pkg.categories if c != 'updates_ignored')
                else:
                    pkg.categories.remove('updates_ignored')
